#
# ddSMT: A delta debugger for SMT benchmarks in SMT-Lib v2 format.
#
# This file is part of ddSMT.
#
# Copyright (C) 2013-2021 by the authors listed in AUTHORS file.
#
# ddSMT is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ddSMT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ddSMT.  If not, see <https://www.gnu.org/licenses/>.

import argparse


class _ModularHelpEnabler(argparse.Action):
    """A custom action that enables help for a single argument group."""
    def __call__(self, parser, namespace, values, option_string=None):
        parser.enable_modular_help(self.const)


class ModularArgumentParser(argparse.ArgumentParser):
    """A variant of argparse.ArgumentParser that allows to modularly enable and
    disable printing help for individual argument groups.

    A new variant of ``add_argument_groups()`` allows adding commands of
    the form ``--help-group``.
    """
    def __init__(self, *args, **kwargs):
        """As ``argparse.ArgumentParser.__init__()``, additionally
        ``modular_action_groups = []`` can be used to specify the names of
        argument groups used for the custom help options."""
        self._modular_action_groups = {}
        self._modular_active = False
        modular_help_groups = kwargs.pop('modular_help_groups', [])
        help_all = kwargs.pop('help_all', True)
        super().__init__(*args, **kwargs)
        self._modular_help_groups = {}
        for name in modular_help_groups:
            self._modular_help_groups[name] = self.add_argument_group(name)
        if help_all:
            self.add_argument('--help-all',
                              action=_ModularHelpEnabler,
                              nargs=0,
                              default=argparse.SUPPRESS,
                              const=None,
                              help='show all help messages')

    def add_argument_group(self, *args, **kwargs):
        """As argparse.ArgumentParser.add_argument_group(), additionally
        help_name, help_group and help_text allow to add a new option
        :code:`--help-{{help_name}}."""
        help_name = kwargs.pop('help_name', None)
        help_group = kwargs.pop('help_group', None)
        help_text = kwargs.pop('help_text', 'show help for {}')
        help_argument = kwargs.pop('help_argument', False)
        grp = super().add_argument_group(*args, **kwargs)
        if help_name is not None:
            self._modular_action_groups[grp] = help_name
            if help_argument:
                parser = self
                if help_group is not None:
                    parser = self._modular_help_groups[help_group]
                parser.add_argument('--help-{}'.format(help_name),
                                    action=_ModularHelpEnabler,
                                    nargs=0,
                                    default=argparse.SUPPRESS,
                                    const=grp,
                                    help=help_text.format(help_name))
        return grp

    def enable_modular_help(self, grp):
        """Remove grp from ``self._modular_action_groups`` and thereby enable
        printing it."""
        self._modular_active = True
        if grp is None:
            self._modular_action_groups = {}
        else:
            del self._modular_action_groups[grp]

    def parse_args(self, *args, **kwargs):
        res = super().parse_args(*args, **kwargs)
        if self._modular_active:
            self.parse_args(['--help'])
        return res

    def error(self, message):
        if not self._modular_active:
            super().error(message)

    def format_help(self):
        """Removes all items from ``self._action_groups`` that are still in
        ``self._modular_action_groups``."""
        self._action_groups = [
            ag for ag in self._action_groups
            if ag not in self._modular_action_groups
        ]
        return super().format_help()



#
# ddSMT: A delta debugger for SMT benchmarks in SMT-Lib v2 format.
#
# This file is part of ddSMT.
#
# Copyright (C) 2013-2021 by the authors listed in AUTHORS file.
#
# ddSMT is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ddSMT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ddSMT.  If not, see <https://www.gnu.org/licenses/>.

from .nodes import Node, count_nodes
from .smtlib import *
from . import options
from .mutator_utils import Simplification


class BinaryReduction:
    """Binary reduction on the top level node.

    This mutator mimics what strategy ``ddmin`` achieves when applying
    mutator :class:`ddsmt.mutators_core.EraseNode`.

    If ``self.ident`` is set, only nodes with the identifier specified as
    ``self.ident`` are considered.

    .. note::
      This mutator is always disabled for strategy ``ddmin``.
    """
    def mutations(self, node):
        if node.is_leaf() or len(node) < 8:
            return
        for sec in nodes.binary_search(len(node)):
            yield Simplification({node.id: node[:sec[0]] + node[sec[1]:]}, [])

    def global_mutations(self, node, input_):
        if node != input_[0]:
            return []
        # generate all sublists as generated by binary-search in bfs order
        # let den be the denominator of the list length (the tree level)
        # let num be the numerator within the current tree level
        if hasattr(self, 'ident'):
            ids = [
                node.id for node in input_
                if is_operator_app(node, self.ident)
            ]
        else:
            ids = [node.id for node in input_]
        for sec in nodes.binary_search(len(ids)):
            yield Simplification(
                {node_id: None
                 for node_id in ids[sec[0]:sec[1]]}, [])

    def __str__(self):
        if hasattr(self, 'ident'):
            return f'binary reduction ({self.ident})'
        return 'binary reduction'


class Constants:
    """Replace given node with a default value of the same sort.

    Default values are defined in :func:`ddsmt.smtlib.get_default_constants`.
    """
    # Requires that global information has been populated via
    # ``collect_information``.
    def filter(self, node):
        return not is_definition_node(node) and get_sort(node) is not None

    def mutations(self, node):
        t = get_sort(node)
        if t is None:
            return []
        res = get_default_constants(t)
        if node in res:
            return []
        for c in res:
            yield Simplification({node.id: c}, [])

    def __str__(self):
        return 'substitute by a constant'


class EraseNode:
    """Erase given given node.

    If ``self.ident`` is set, only nodes with the identifier specified as
    ``self.ident`` are considered.
    """
    def filter(self, node):
        if not hasattr(self, 'ident'):
            return True
        return node.has_ident() and node.get_ident() == self.ident

    def mutations(self, node):
        return [Simplification({node.id: None}, [])]

    def __str__(self):
        if hasattr(self, 'ident'):
            return f'erase node ({self.ident})'
        return 'erase node'


class MergeWithChildren:
    """Merge given node with one of its children.

    This mutator can only be applied to n-ary operations like ``and`` or ``+``.
    """
    def filter(self, node):
        return has_nary_operator(node)

    def mutations(self, node):
        for cid, child in enumerate(node):
            if child.has_ident() and node.get_ident() == child.get_ident():
                simp = Node(*node[:cid], *node[cid][1:], *node[cid + 1:])
                yield Simplification({node.id: simp}, [])

    def __str__(self):
        return 'merge with child'


class ReplaceByChild:
    """Replace given node with one of its children."""
    def filter(self, node):
        return not is_leaf(node) and not is_operator_app(node, 'let')

    def mutations(self, node):
        sort = get_sort(node)
        for n in node[1:]:
            if get_sort(n) == sort:
                yield Simplification({node.id: n}, [])

    def __str__(self):
        return 'replace by a child'


class ReplaceByVariable:
    """Replace given node with an existing variable of the same type.

    .. note::
      Replacing variables with other variables **potentially introduces
      cycles**.
      To avoid this, for leaf nodes we only allow substitutions with
      variables that are lexicographically **smaller** (default)
      or **larger** than the leaf mode.
      Configure with option ``--replace-by-variable-mode`` (use ``inc`` for
      substitution with smaller, and ``dec`` for substition with larger
      variables).
    """
    # Requires that global information has been populated via
    # ``collect_information``.
    def filter(self, node):
        return not is_const(node) and not is_definition_node(
            node) and get_sort(node) is not None

    def mutations(self, node):
        if not hasattr(self, 'repl_mode'):
            self.repl_mode = options.args().replace_by_variable_mode

        ret_sort = get_sort(node)
        if ret_sort is None:
            return []
        variables = get_variables_with_sort(ret_sort)
        # Replacing by a defined variable may loop with inlining
        variables = filter(lambda v: not is_defined_fun(v), variables)
        if is_leaf(node):
            if self.repl_mode == 'inc':
                variables = [v for v in variables if v > node.data]
            else:
                variables = [v for v in variables if v < node.data]
        yield from [Simplification({node.id: Node(v)}, []) for v in variables]

    def __str__(self):
        if hasattr(self, 'repl_mode'):
            return f'substitute by existing variable ({self.repl_mode})'
        return 'substitute by existing variable'


class SortChildren:
    """Sort the children of a given node by size (count of sub-nodes)."""
    def filter(self, node):
        return not is_leaf(node)

    def mutations(self, node):
        s = nodes.Node(*sorted(node, key=count_nodes))
        if s != node:
            return [Simplification({node.id: s}, [])]
        return []

    def __str__(self):
        return 'sort children'


def get_mutator_options(argparser):
    """Add additional options to the argument parser."""
    argparser.add_argument(
        '--replace-by-variable-mode',
        choices=['inc', 'dec'],
        default='inc',
        help='replace with existing variables that are larger or smaller')


def get_mutators():
    """Return a mapping from mutator class names to the name of their config
    options."""
    return {
        'BinaryReduction': 'binary-reduction',
        'Constants': 'constants',
        'EraseNode': 'erase-node',
        'MergeWithChildren': 'merge-children',
        'ReplaceByChild': 'substitute-children',
        'ReplaceByVariable': 'replace-by-variable',
        'SortChildren': 'sort-children',
    }



#
# ddSMT: A delta debugger for SMT benchmarks in SMT-Lib v2 format.
#
# This file is part of ddSMT.
#
# Copyright (C) 2013-2021 by the authors listed in AUTHORS file.
#
# ddSMT is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ddSMT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ddSMT.  If not, see <https://www.gnu.org/licenses/>.

from .smtlib import *
from .mutator_utils import Simplification


def is_quantifier(node):
    return node.has_ident() and node.get_ident() in ['exists', 'forall']


def make_and(children):
    """Return an and from arbitrary many children.

    Returns node as a list so that we can return the empty list if there
    are no children.
    """
    if len(children) == 0:
        return None
    if len(children) == 1:
        return children[0]
    return Node('and', *children)


class BoolDeMorgan:
    """Use de Morgans rules to push negations inside."""
    def filter(self, node):
        return is_operator_app(node, 'not') and len(node) > 1 and (
            is_operator_app(node[1], 'and') or is_operator_app(node[1], 'or'))

    def mutations(self, node):
        negated = [Node('not', t) for t in node[1][1:]]
        if node[1].get_ident() == 'and':
            return [Simplification({node.id: Node('or', *negated)}, [])]
        assert node[1].get_ident() == 'or'
        return [Simplification({node.id: Node('and', *negated)}, [])]

    def __str__(self):
        return 'push negation inside'


class BoolDoubleNegation:
    """Eliminate double negations."""
    def filter(self, node):
        return is_operator_app(
            node, 'not') and len(node) > 1 and is_operator_app(node[1], 'not')

    def mutations(self, node):
        return [Simplification({node.id: node[1][1]}, [])]

    def __str__(self):
        return 'eliminate double negation'


class BoolEliminateFalseEquality:
    """Replace an equality with ``false`` by a negation.

    For example, transform ``(= false X)`` into ``(not X)``.
    """
    def filter(self, node):
        return is_eq(node) and Node('false') in node

    def mutations(self, node):
        simp = make_and([Node('not', n) for n in node[1:] if n != 'false'])
        return [] if simp is None else [Simplification({node.id: simp}, [])]

    def __str__(self):
        return 'replace equality with false by negation'


class BoolEliminateImplication:
    """Replace (possibly n-ary) implications by disjunctions."""
    def filter(self, node):
        return is_operator_app(node, '=>')

    def mutations(self, node):
        simp = make_and([
            Node('or', ('not', node[i]), node[i + 1])
            for i in range(1,
                           len(node) - 1)
        ])
        return [] if simp is None else [Simplification({node.id: simp}, [])]

    def __str__(self):
        return 'eliminate implication'


class BoolNegateQuantifier:
    """Push negations inside quantifiers."""
    def filter(self, node):
        return is_operator_app(
            node, 'not') and len(node) > 1 and is_quantifier(node[1])

    def mutations(self, node):
        if node[1].get_ident() == 'exists':
            return [
                Simplification(
                    {node.id: Node('forall', node[1][1], ('not', node[1][2]))},
                    [])
            ]
        assert node[1].get_ident() == 'forall'
        return [
            Simplification(
                {node.id: Node('exists', node[1][1], ('not', node[1][2]))}, [])
        ]

    def __str__(self):
        return 'push negation inside of quantifier'


class BoolXOREliminateBinary:
    """Eliminate binary ``xor`` by ``distinct``."""
    def filter(self, node):
        return is_operator_app(node, 'xor') and len(node) == 3

    def mutations(self, node):
        return [
            Simplification({node.id: Node('distinct', node[1], node[2])}, [])
        ]

    def __str__(self):
        return 'eliminate binary xor'


class BoolXORRemoveConstant:
    """Eliminate constant children from ``xor``."""
    def filter(self, node):
        return is_operator_app(node, 'xor')

    def mutations(self, node):
        if 'false' in node:
            yield Simplification(
                {node.id: Node(*[c for c in node if c != 'false'])}, [])
        if 'true' in node:
            yield Simplification(
                {node.id: Node(*[c for c in node if c != 'true'])}, [])
            yield Simplification(
                {node.id: Node('not', tuple(c for c in node if c != 'true'))},
                [])

    def __str__(self):
        return 'remove constants from xor'


def get_mutators():
    """Returns a mapping from mutator class names to the name of their config
    options."""
    return {
        'BoolDeMorgan': 'bool-de-morgan',
        'BoolDoubleNegation': 'bool-double-negations',
        'BoolEliminateFalseEquality': 'bool-false-eq',
        'BoolEliminateImplication': 'bool-implications',
        'BoolNegateQuantifier': 'bool-negate-quants',
        'BoolXOREliminateBinary': 'bool-xor-binary',
        'BoolXORRemoveConstant': 'bool-xor-constants',
    }



#
# ddSMT: A delta debugger for SMT benchmarks in SMT-Lib v2 format.
#
# This file is part of ddSMT.
#
# Copyright (C) 2013-2021 by the authors listed in AUTHORS file.
#
# ddSMT is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ddSMT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ddSMT.  If not, see <https://www.gnu.org/licenses/>.

import argparse
import logging

from . import mutators_arithmetic
from . import mutators_bv
from . import mutators_boolean
from . import mutators_core
from . import mutators_fp
from . import mutators_smtlib
from . import mutators_strings
from . import nodes
from . import options


def get_all_mutators():
    """Return all available mutators, arranged by their theory."""
    return {
        'core': (mutators_core, mutators_core.get_mutators()),
        'arithmetic':
        (mutators_arithmetic, mutators_arithmetic.get_mutators()),
        'bv': (mutators_bv, mutators_bv.get_mutators()),
        'boolean': (mutators_boolean, mutators_boolean.get_mutators()),
        'fp': (mutators_fp, mutators_fp.get_mutators()),
        'smtlib': (mutators_smtlib, mutators_smtlib.get_mutators()),
        'strings': (mutators_strings, mutators_strings.get_mutators()),
    }


def get_mutators(mutators):
    """Return mutator instances from a list of names.

    For each mutator class name in ``mutators`` retrieves the proper
    theory, checks whether the mutator is enabled and then adds an
    instance of this class to the result.
    """
    res = []
    for m in mutators:
        for _, theory in get_all_mutators().items():
            # check if this mutator belongs to this theory
            if m in theory[1]:
                attr = f'mutator_{theory[1][m].replace("-", "_")}'
                # check if this mutator is enabled
                if getattr(options.args(), attr, True):
                    res.append(getattr(theory[0], m)())
                break
    return res


def get_initialized_mutator(name, properties):
    """Obtain a mutator by name and set the given properties."""
    res = get_mutators([name])
    if res:
        for p, v in properties.items():
            setattr(res[0], p, v)
    return res


def toggle_theory(namespace, theory_name, value):
    """Enables or disables all mutators for the given theory by setting their
    respective options in ``namespace``."""
    setattr(namespace, f'mutators_{theory_name}', value)
    _, mutators = get_all_mutators()[theory_name]
    for _, opt in mutators.items():
        setattr(namespace, f'mutator_{opt.replace("-", "_")}', value)


def toggle_all_theories(namespace, value):
    """Enables or disables all mutators for all theories by setting their
    respective options in ``namespace``."""
    for theory_name, data in get_all_mutators().items():
        setattr(namespace, f'mutators_{theory_name}', value)
        for _, opt in data[1].items():
            setattr(namespace, f'mutator_{opt.replace("-", "_")}', value)


class TheoryToggleAction(options.ToggleAction):
    """A specialization of ``ToggleAction`` that disables (or enables) all
    mutators of a given theory."""
    def __init__(self, theory, *args, **kwargs):
        """Expects an additional option with the name of the theory."""
        self.__theory = theory
        super(TheoryToggleAction, self).__init__(*args, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        """Set the option, as well as all mutator options for this theory."""
        value = self._get_value(option_string)
        setattr(namespace, self.dest, value)
        toggle_theory(namespace, self.__theory, value)


class DisableAllTheoriesAction(argparse.Action):
    """Disables all mutators from all theories when called."""
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, True)
        toggle_all_theories(namespace, False)


def add_mutator_group(argparser, name):
    """Add a new argument group for a mutator group, including options to
    enable and disable the whole theory."""
    res = argparser.add_argument_group('{} mutator arguments'.format(name),
                                       help_name=name,
                                       help_group='mutator help',
                                       help_text='show help for {} mutators')

    res._add_action(
        TheoryToggleAction(name,
                           name,
                           default=None,
                           dest=f'mutators_{name}',
                           help=f'{name} theory'))
    return res


def collect_mutator_options(argparser):
    """Adds all options related to mutators to the given argument parser."""
    argparser.add_argument('--disable-all',
                           action=DisableAllTheoriesAction,
                           nargs=0,
                           help='disable all mutators')
    for name, tdata in get_all_mutators().items():
        theory = tdata[0]
        # add argument group for this theory
        ap = add_mutator_group(argparser, name)
        for mname, mopt in theory.get_mutators().items():
            # add options for every individual mutator
            mdesc = str(getattr(theory, mname)())
            ap._add_action(
                options.ToggleAction(mopt,
                                     dest=f'mutator_{mopt.replace("-", "_")}',
                                     help=mdesc))
        # add custom option, if the theory wants it
        if hasattr(theory, 'get_mutator_options'):
            theory.get_mutator_options(ap)


def auto_detect_theories(exprs):
    for name, tdata in get_all_mutators().items():
        # if a theory was explicitly enabled or disabled by the user, we don't
        # change it.
        optval = getattr(options.args(), f'mutators_{name}')
        if optval is not None:
            logging.debug(f'{name} was specified by the user: {optval}')
            continue

        theory = tdata[0]
        # if a theory does not implement is_relevant(), it stays enabled
        if not hasattr(theory, 'is_relevant'):
            logging.debug(f'{name} can not be disabled')
            continue

        # now check whether theory is relevant for any of the nodes
        enabled = False
        for node in nodes.dfs(exprs, max_depth=1):
            if theory.is_relevant(node):
                enabled = True
                break

        if not enabled:
            logging.warn(f'automatically disabling {name} mutators. '
                         f'Use --{name} to use it.')
            toggle_theory(options.args(), name, False)
        else:
            logging.debug(f'keeping {name} mutators enabled')



#
# ddSMT: A delta debugger for SMT benchmarks in SMT-Lib v2 format.
#
# This file is part of ddSMT.
#
# Copyright (C) 2013-2021 by the authors listed in AUTHORS file.
#
# ddSMT is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ddSMT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ddSMT.  If not, see <https://www.gnu.org/licenses/>.

import collections
import multiprocessing
import logging
import pickle
import sys
import time
import traceback

from . import checker
from . import mutators
from . import nodeio
from . import nodes
from . import options
from . import debug_utils
from . import smtlib
from .mutator_utils import Simplification, apply_simp

Task = collections.namedtuple('Task', ['id', 'exprs', 'simplifications'])

Result = collections.namedtuple(
    'Result', ['task_id', 'success', 'reduced', 'exprs', 'tests'])


def _partition(exprs, gran):
    """Partition ``exprs`` into a list of subsets of size ``gran``."""
    return [exprs[s:s + gran] for s in range(0, len(exprs), gran)]


class TaskGenerator:
    """Filter ``exprs`` based on ``mutator`` and generates tasks."""
    def __init__(self, exprs, gran, mutator, max_depth=None):
        self.exprs = exprs
        self.mutator = mutator
        self.max_depth = max_depth
        self.index = 0
        self.stopped = False

        # Filter nodes and partition into subsets of size ``gran``.
        filter_func = getattr(mutator, 'filter', lambda x: True)
        filtered = list(nodes.filter_nodes(exprs, filter_func, max_depth))
        self.num_filtered = len(filtered)
        self.gran = len(filtered) if gran is None else gran
        self.subsets = _partition(filtered, self.gran) if self.gran else []

        njobs = options.args().jobs
        if njobs > 1 and len(self.subsets) > 2 * njobs:
            self.pickled_exprs = pickle.dumps(exprs)
        else:
            self.pickled_exprs = None

    def __iter__(self):
        return self

    def __next__(self):
        """Generate next task."""
        while not self.stopped and self.index < len(self.subsets):
            task_id = self.index
            self.index += 1

            # Filter nodes in subset in order to ensure that the mutator still
            # applies after updating ``self.exprs`` via ``self.update``.
            subset = self.subsets[task_id]
            subset = [n for n in subset if self.mutator.filter(n)]
            if not subset:
                continue

            simps = self.__get_substs(subset)

            if not simps:
                continue

            logging.debug(f'TaskGen: Generate next {task_id}')
            if self.pickled_exprs:
                return Task(task_id, self.pickled_exprs, pickle.dumps(simps))
            return Task(task_id, self.exprs, simps)
        raise StopIteration

    def __get_substs(self, subset):
        """Generate substitutions for ``subset`` based on ``self.mutator``."""
        # Granularity 1: Try all mutations separately.
        if len(subset) == 1:
            node = subset[0]
            if hasattr(self.mutator, 'mutations'):
                mutations = self.mutator.mutations(node)
            elif hasattr(self.mutator, 'global_mutations'):
                mutations = self.mutator.global_mutations(node, self.exprs)
            else:
                return None
            return list(mutations)

        # Granularity > 1: Pick first simplification for each node and group
        # them all into one.
        fresh_vars = []
        substs = dict()
        for node in subset:
            if hasattr(self.mutator, 'mutations'):
                mutations = self.mutator.mutations(node)
            elif hasattr(self.mutator, 'global_mutations'):
                mutations = self.mutator.global_mutations(node, self.exprs)
            else:
                continue

            try:
                simp = next(iter(mutations))
                assert isinstance(simp, Simplification)
                fresh_vars.extend(simp.fresh_vars)
                substs.update(simp.substs)
            except StopIteration:
                continue

        if not substs:
            return None
        return [Simplification(substs, fresh_vars)]

    def reset(self, index):
        """Reset ``self.index`` to ``index``."""
        logging.debug(f'TaskGen: Reset to {index}')
        self.index = index

    def update(self, exprs):
        """Update ``self.exprs`` with new ``exprs``."""
        self.exprs = exprs
        if self.pickled_exprs is not None:
            self.pickled_exprs = pickle.dumps(exprs)

    def stop(self):
        """Stop generating new taks."""
        self.stopped = True

    def start(self):
        """Continue generating new taks."""
        self.stopped = False


def ddmin_passes():
    """Return list of ddmin passes.

    Mutators that apply global_mutations only are disabled since this is
    not supported by ddmin.
    """

    # Passes applied to top-level nodes (DFS max-depth 1)
    stage1_names = ['EraseNode', 'CheckSatAssuming']

    stage1 = mutators.get_initialized_mutator('EraseNode', {'ident': 'assert'})
    stage1.extend(mutators.get_mutators(stage1_names))

    # Use mutators that promise maximum reduction first
    stage2_names = [
        'Constants',
        'ReplaceByChild',
        'IntroduceFreshVariable',
        'LetElimination',
        'LetSubstitution',
        'ReplaceByVariable',
        'MergeWithChildren',
    ]

    exclude = [
        'BinaryReduction',
    ]
    exclude.extend(stage1_names)
    exclude.extend(stage2_names)

    # Add remaining mutators
    for theory in mutators.get_all_mutators().values():
        stage2_names.extend(x for x in theory[1] if x not in exclude)

    return [stage1, mutators.get_mutators(stage2_names)]


def _simp(exprs, simplifications):
    """Apply ``simplifications`` to ``exprs`` and return simplified
    formulas."""
    for simp in simplifications:
        mexprs = apply_simp(exprs, simp)
        if mexprs is not None:
            yield mexprs


__cached_exprs = None
__cached_exprs_hash = None
__abort_flag = None


def _worker(task):
    """Process given ``task``.

    If _worker runs in a separate process ``task.exprs`` and
    ``task.simplifications`` are pickled and need to be unpickled before
    performing the substitutions and checks.
    """
    global __cached_exprs
    global __cached_exprs_hash
    global __abort_flag

    with debug_utils.Profiler():
        try:
            if __abort_flag and __abort_flag.is_set():
                logging.debug(f'Worker: Abort task {task.id}')
                return Result(task.id, False, 0, [], 0)

            if isinstance(task.exprs, bytes):
                hashval = hash(task.exprs)
                if __cached_exprs_hash != hashval:
                    __cached_exprs = pickle.loads(task.exprs)
                    __cached_exprs_hash = hashval

                exprs = __cached_exprs
                substs = pickle.loads(task.simplifications)
            else:
                exprs = task.exprs
                substs = task.simplifications

            ntests = 0
            for mexprs in _simp(exprs, substs):
                ntests += 1
                if checker.check_exprs(mexprs):
                    nreduced = (nodes.count_exprs(exprs)
                                - nodes.count_exprs(mexprs))
                    return Result(task.id, True, nreduced, mexprs, ntests)
            return Result(task.id, False, 0, [], ntests)
        except Exception as e:
            logging.info(f'{type(e)} in ddmin worker: {e}')
            exc_type, exc_value, exc_traceback = sys.exc_info()
            traceback.print_tb(exc_traceback, limit=10, file=sys.stderr)


__last_msg = ""


def _print_progress(msg, update=True):
    """Print progress ``msg``.

    If ``update`` is True the previous line will be overwritten.
    """
    global __last_msg

    if options.args().verbosity < 1:
        return

    if sys.stderr.isatty():
        print(' ' * len(__last_msg), end='\r', flush=True, file=sys.stderr)
    if update and sys.stderr.isatty():
        __last_msg = f'[ddSMT INFO] {msg}'
        print(__last_msg, end='\r', flush=True, file=sys.stderr)
    else:
        logging.info(msg)
        __last_msg = msg


def _check_seq(taskgen, nexprs, stats):
    """Sequentially process tasks generated by ``taskgen``."""

    outfile = options.args().outfile

    for task in taskgen:
        result = _worker(task)
        stats['tests'] += result.tests

        if result.success:
            stats['tests_success'] += 1
            stats['reduced'] += result.reduced
            taskgen.update(result.exprs)
            nodeio.write_smtlib_to_file(outfile, taskgen.exprs)
            smtlib.collect_information(taskgen.exprs)

        _print_progress(
            f"{taskgen.mutator}: "
            f"nodes: {taskgen.num_filtered}, "
            f"gran: {taskgen.gran}, "
            f"subset {task.id} of {len(taskgen.subsets)}, "
            f"exprs: {nexprs - stats['reduced']}/{nexprs}",
            options.args().verbosity == 1)

    return taskgen.exprs


def _check_par(taskgen, nexprs, stats):
    """Process tasks generated by ``taskgen`` with multiple processes.

    As soon as one process performs a successful check with task N the
    main process stops the task generator and notifies all worker
    processes to stop ASAP. ``taskgen.exprs`` will be updated with the
    reduced expressions and ``taskgen`` is reset to start with task N+1.
    """
    global __abort_flag

    outfile = options.args().outfile

    __abort_flag = multiprocessing.Manager().Event()

    start_index = 0
    with multiprocessing.Pool(options.args().jobs) as pool:
        while start_index >= 0:
            start_index = -1
            skip = False
            for result in pool.imap_unordered(_worker, taskgen):
                stats['tests'] += result.tests

                if result.success and not skip:
                    __abort_flag.set()
                    logging.debug('Main: Set abort flag')
                    taskgen.stop()
                    taskgen.update(result.exprs)
                    nodeio.write_smtlib_to_file(outfile, taskgen.exprs)
                    stats['tests_success'] += 1
                    stats['reduced'] += result.reduced
                    start_index = result.task_id + 1
                    skip = True
                    logging.debug(
                        f'Successful test with subset {result.task_id}: '
                        f'{result.reduced}')
                elif result.success and skip:
                    logging.debug(f'Skip test with subset {result.task_id}: '
                                  f'{result.reduced}')

                _print_progress(
                    f"{taskgen.mutator}: "
                    f"nodes: {taskgen.num_filtered}, "
                    f"gran: {taskgen.gran}, "
                    f"subset {result.task_id} of {len(taskgen.subsets)}, "
                    f"exprs: {nexprs - stats['reduced']}/{nexprs}",
                    options.args().verbosity == 1)

            if __abort_flag.is_set():
                __abort_flag.clear()
                smtlib.collect_information(taskgen.exprs)
                taskgen.reset(start_index)
                taskgen.start()
                logging.debug(f'Restart tests starting from {start_index}')

    return taskgen.exprs


def _apply_mutator(mutator, exprs, max_depth=None):
    """Apply ``mutator`` with strategy ddmin on input ``exprs``.

    ``max_depth`` limits the DFS traversal when filtering nodes in
    ``exprs``.
    """

    start_time = time.time()
    nexprs = nodes.count_exprs(exprs)
    stats = {'tests': 0, 'tests_success': 0, 'reduced': 0}
    taskgen = TaskGenerator(exprs, None, mutator, max_depth)
    gran = taskgen.gran
    while gran > 0:
        check_func = _check_par if taskgen.pickled_exprs else _check_seq
        exprs = check_func(taskgen, nexprs, stats)
        exprs = nodes.reduplicate(exprs)
        gran = gran // 2
        taskgen = TaskGenerator(exprs, gran, mutator, max_depth)

    if stats['tests'] > 0 or options.args().verbosity >= 2:
        _print_progress(
            f"{mutator}: diff {-stats['reduced']:+} exprs, "
            f"{stats['tests']} tests ({stats['tests_success']}), "
            f"{time.time() - start_time:.1f}s", False)

    return exprs, stats['tests'], stats['reduced']


def reduce(exprs):
    """Reduce given ``exprs`` until fixed-point with ddmin strategy."""

    smtlib.collect_information(exprs)

    passes = ddmin_passes()

    ntests_total = 0
    loop_checker = debug_utils.NodeLoopChecker()

    while True:
        loop_checker.add(exprs)
        nreduced_round = 0

        # Apply top-level passes until fixed-point.
        for mut in passes[0]:
            while True:
                exprs, ntests, nreduced = _apply_mutator(mut, exprs, 1)
                ntests_total += ntests
                nreduced_round += nreduced

                if nreduced == 0:
                    break

        for mut in passes[1]:
            exprs, ntests, nreduced = _apply_mutator(mut, exprs)
            ntests_total += ntests
            nreduced_round += nreduced

        if nreduced_round == 0:
            break

    return exprs, ntests_total



#
# ddSMT: A delta debugger for SMT benchmarks in SMT-Lib v2 format.
#
# This file is part of ddSMT.
#
# Copyright (C) 2013-2021 by the authors listed in AUTHORS file.
#
# ddSMT is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ddSMT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ddSMT.  If not, see <https://www.gnu.org/licenses/>.

from . import nodes
from .nodes import Node
from .smtlib import *
from .mutator_utils import Simplification


class BVNormalizeConstants:
    """Normalize bit-vector constants to be represented using the ``(_ bvN M)``
    notation."""
    def filter(self, node):
        return node.is_leaf() and is_bv_const(node)

    def mutations(self, node):
        val, bw = get_bv_constant_value(node)
        return [Simplification({node.id: Node('_', f'bv{val}', bw)}, [])]

    def __str__(self):
        return 'normalize bit-vector const'


class BVConcatToZeroExtend:
    """Replace a ``concat`` with zero by ``zero_extend``."""
    def filter(self, node):
        if not node.has_ident() or node.get_ident() != 'concat':
            return False
        if not is_bv_const(node[1]):
            return False
        return get_bv_constant_value(node[1])[0] == 0

    def mutations(self, node):
        return [
            Simplification(
                {
                    node.id:
                    Node(('_', 'zero_extend', get_bv_constant_value(
                        node[1])[1]), node[2])
                }, [])
        ]

    def __str__(self):
        return 'replace concat by zero_extend'


class BVDoubleNegation:
    """Eliminate double bit-vector negations."""
    def filter(self, node):
        if is_bv_not(node) and len(node) > 1 and is_bv_not(node[1]):
            return True
        if is_bv_neg(node) and is_bv_neg(node[1]):
            return True
        return False

    def mutations(self, node):
        return [Simplification({node.id: node[1][1]}, [])]

    def __str__(self):
        return 'eliminate double bit-vector negation'


class BVElimBVComp:
    """Replace equality over ``bvcomp`` and constant with a regular equality.

    For example, ``(= (bvcomp a b) #b1)`` is replaced with ``(= a b)``.
    """
    def filter(self, node):
        return (is_eq(node) and len(node) > 2 and is_bv_const(node[1])
                and get_bv_width(node[1]) == 1
                and any(is_bv_comp(n) for n in node[2:]))

    def mutations(self, node):
        val = get_bv_constant_value(node[1])[0]
        res = []
        for n in node[2:]:
            if is_bv_comp(n):
                if val == 1:
                    res.append(Node('=', *n[1:]))
                else:
                    res.append(Node('not', Node('=', *n[1:])))
            else:
                res.append(Node('=', node[1], n))
        if len(res) == 1:
            return [Simplification({node.id: res[0]}, [])]
        else:
            return [Simplification({node.id: Node('and', *res)}, [])]

    def __str__(self):
        return 'eliminate bvcomp by equality'


class BVEvalExtend:
    """Evaluate a bit-vector ``(sign|zero)_extend`` if it is applied to a
    constant or another ``(sign|zero)_extend``."""
    def filter(self, node):
        return ((is_indexed_operator_app(node, 'zero_extend')
                 or is_indexed_operator_app(node, 'sign_extend'))
                and len(node) > 1 and is_bv_const(node[1]))

    def mutations(self, node):
        (val, width) = get_bv_constant_value(node[1])
        index = get_indices(node[0], node[0][1])[0]
        if is_indexed_operator_app(node, 'sign_extend'):
            val_bin = bin(val)
            if len(val_bin[2:]) == width and val_bin[2] == '1':
                ones = '1' * index
                return [
                    Simplification({node.id: Node(f'#b{ones}{val_bin[2:]}')},
                                   [])
                ]
        return [
            Simplification({node.id: Node('_', f'bv{val}', width + index)}, [])
        ]

    def __str__(self):
        return 'evaluate bit-vector extend'


class BVExtractConstants:
    """Evaluate a bit-vector ``extract`` if it is applied to a constant."""
    def filter(self, node):
        return (is_indexed_operator_app(node, 'extract', 2) and len(node) > 1
                and is_bv_const(node[1]))

    def mutations(self, node):
        if node[1].has_ident():
            assert node[1][0] == '_'
            val, bw = get_bv_constant_value(node[1])
            constant = bin(val)[2:]
            diff = bw - len(constant)
            if diff > 0:
                constant = f"{'0' * diff}{constant}"
        else:
            assert node[1].data.startswith('#b')
            constant = node[1][2:]
        n = len(constant)
        idx = get_indices(node[0], 'extract', 2)
        upper = n - idx[0] - 1
        lower = n - idx[1] - 1
        return [
            Simplification({node.id: Node(f'#b{constant[upper:lower+1]}')}, [])
        ]

    def __str__(self):
        return 'evaluate bit-vector extract on constant'


class BVExtractZeroExtend:
    """Simplify an ``extract`` of a ``zero_extend`` by pushing the
    ``zero_extend`` to the outside and reducing the bit-width, if possible.
    """
    #Requires that global information has been populated via
    #``collect_information``.
    def filter(self, node):
        return (is_indexed_operator_app(node, 'extract', 2) and len(node) > 1
                and is_indexed_operator_app(node[1], 'zero_extend'))

    def mutations(self, node):
        term = node[1][1]
        bw = get_bv_width(term)
        if bw <= 0:
            return []
        idx = get_indices(node[0], 'extract', 2)
        upper = idx[0]
        lower = idx[1]
        # we extract upper..lower from 0|varwidth
        if lower >= bw:
            # we only extract from the zeroes
            return [
                Simplification({node.id: Node('_', 'bv0', upper - lower + 1)},
                               [])
            ]
        if upper < bw:
            # we only extract from the variable
            return [Simplification({node.id: Node(node[0], term)}, [])]
        # switch extract and zero_extend, reduce lengths of extract and
        # zero_extend
        return [
            Simplification(
                {
                    node.id:
                    Node(('_', 'zero_extend', str(upper - bw + 1)),
                         (('_', 'extract', str(bw - 1), str(lower)), term))
                }, [])
        ]

    def __str__(self):
        return 'simplify bit-vector extract on zero_extend'


class BVIteToBVComp:
    """Replace an ``ite`` of bit-width one with ``bvcomp``."""
    #Requires that global information has been populated via
    #``collect_information``.
    def filter(self, node):
        if not is_operator_app(node, 'ite'):
            return False
        if (not is_operator_app(node[1], '=') or len(node[1]) != 3
                or not is_bv_sort(get_sort(node[1][1]))
                or get_sort(node[1][1]) is None):
            return False
        if not is_bv_const(node[2]) or not is_bv_const(node[3]):
            return False
        if get_bv_constant_value(node[2]) != (1, 1):
            return False
        if get_bv_constant_value(node[3]) != (0, 1):
            return False
        return True

    def mutations(self, node):
        return [
            Simplification({node.id: Node('bvcomp', node[1][1], node[1][2])},
                           [])
        ]

    def __str__(self):
        return 'eliminate ite with bv1 / bv0 cases'


class BVReflexiveNand:
    """Replace a reflexive ``bvnand`` by bitwise negation."""
    def filter(self, node):
        return (is_operator_app(node, 'bvnand') and len(node) == 3
                and node[1] == node[2])

    def mutations(self, node):
        return [Simplification({node.id: Node('bvnot', node[1])}, [])]

    def __str__(self):
        return 'replace bvnand by bvnot'


class BVSimplifyConstants:
    """Replace a bit-vector constant with a simpler constant of smaller value.
    """
    def filter(self, node):
        return (is_bv_const(node)
                and get_bv_constant_value(node)[0] not in [0, 1])

    def mutations(self, node):
        val, width = get_bv_constant_value(node)
        yield Simplification(
            {node.id: Node('#b{{:0>{}b}}'.format(width).format(0))}, [])
        yield Simplification(
            {node.id: Node('#b{{:0>{}b}}'.format(width).format(1))}, [])
        yield from [
            Simplification(
                {node.id: Node('#b{{:0>{}b}}'.format(width).format(v))}, [])
            for v in [val // 32, val // 8, val // 2] if v not in [0, 1]
        ]

    def global_mutations(self, node, input_):
        yield from [
            Simplification({node: simp.substs[node.id]}, [])
            for simp in self.mutations(node)
        ]

    def __str__(self):
        return 'simplify bit-vector constant'


class BVTransformToBool:
    """Turn bit-vector operators into Boolean operators.

    For example, transform ``(= bv1 (bvor X Y))`` into ``(or (= bv1 X)
    (= bv1 Y))``.
    """
    repl = {'bvand': 'and', 'bvor': 'or', 'bvxor': 'xor'}

    def filter(self, node):
        if not is_operator_app(node, '=') or len(node) != 3:
            return False

        if is_bv_const(node[1]) and get_bv_width(node[1]) == 1:
            return node[2].has_ident() and node[2].get_ident() in self.repl

        return (is_bv_const(node[2]) and get_bv_width(node[2]) == 1
                and node[1].has_ident() and node[1].get_ident() in self.repl)

    def mutations(self, node):
        if is_bv_const(node[1]) and get_bv_width(node[1]) == 1:
            c = node[1]
            n = node[2]
        else:
            c = node[2]
            n = node[1]
        return [
            Simplification(
                {
                    node.id:
                    Node(self.repl[n.get_ident()], *
                         [Node('=', c, d) for d in n[1:]])
                }, [])
        ]

    def __str__(self):
        return 'transform bit-vector to boolean'


class BVZeroExtendPredicate():
    """Eliminate ``zero_extend`` when both operands of an equality, disequality
    or inequality are zero extended.

    For example, transform ``(= ((_ zero_extend 2) a) ((_ zero_extend 2) b))``
    into ``(= a b)``, and ``(= ((_ zero_extend 2) a) ((_ zero_extend 4) b))``
    into ``(= a ((_ zero_extend 2 b)))``.
    """
    def filter(self, node):
        return node.has_ident() and len(node) == 3 and node.get_ident() in [
            '=', 'distinct', 'bvult', 'bvule', 'bvugt', 'bvuge', 'bvslt',
            'bvsle', 'bvsgt', 'bvsge'
        ] and is_indexed_operator_app(node[1], 'zero_extend',
                                      1) and is_indexed_operator_app(
                                          node[2], 'zero_extend', 1)

    def mutations(self, node):
        idx1 = get_indices(node[1][0], 'zero_extend', 1)[0]
        idx2 = get_indices(node[2][0], 'zero_extend', 1)[0]
        if idx1 == idx2:
            return [
                Simplification(
                    {node.id: Node(node.get_ident(), node[1][1], node[2][1])},
                    [])
            ]
        if idx1 > idx2:
            return [
                Simplification(
                    {
                        node.id:
                        Node(
                            node.get_ident(),
                            Node(('_', 'zero_extend', int(idx1) - int(idx2)),
                                 node[1][1]), node[2][1])
                    }, [])
            ]
        return [
            Simplification(
                {
                    node.id:
                    Node(
                        node.get_ident(), node[1][1],
                        Node(('_', 'zero_extend', int(idx2) - int(idx1)),
                             node[2][1]))
                }, [])
        ]

    def __str__(self):
        return 'eliminate zero_extend when both operands of a predicate are \
                zero extended'


class BVReduceBW:
    """Reduce the bit-width of a variable by introducing an ``extract`` and
    ``zero_extend`` on that variable.

    For example, ``(declare-const v (_ BitVec 32))`` is transformed into

    .. code-block:: common-lisp

        (declare-const _v (_BitVec 1))
        (define-fun v () (_ BitVec 32) ((_ zero_extend 31) _v))
    """
    #Requires that global information has been populated via
    #``collect_information``.
    def filter(self, node):
        if not node.has_ident():
            return False
        print(node)
        print(len(node))
        return (((node.get_ident() == 'declare-const' and len(node) > 2) or
                 (node.get_ident() == 'declare-fun' and len(node) > 3
                  and len(node[2]) == 0)) and get_sort(node[1]) is not None
                and is_bv_sort(get_sort(node[1])))

    def global_mutations(self, node, input_):
        bw = get_bv_width(node[1])
        bws = sorted(set([bw - 1, bw // 2, 2, 1]))
        for b in bws:
            if 0 < b < bw:
                varname = '_{}'.format(node[1])
                var = Node('declare-const', varname, Node('_', 'BitVec', b))
                zext = Node('define-fun', node[1], (), get_sort(node[1]),
                            Node(Node('_', 'zero_extend', bw - b), varname))
                yield Simplification({node.id: zext}, [var])

    def __str__(self):
        return 'reduce bit-width of variable'


class BVMergeReducedBW:
    """Merge previous bit-width reductions of the form.

    .. code-block:: common-lisp

        (declare-const __w (_BitVec MM))
        (define-fun _w () (_ BitVec Y) ((_ zero_extend N) __w))
        (define-fun w () (_ BitVec X) ((_ zero_extend M) _w))``

    into

    .. code-block:: common-lisp

        (declare-const __w (_BitVec MM))
        (define-fun _w () (_ BitVec Y) ((_ zero_extend N) __w))
        (define-fun w () (_ BitVec X) ((_ zero_extend M+N) __w))

    Obsolete ``define-fun`` expressions will be removed later on.
    """
    def filter(self, node):
        return (is_operator_app(node, 'define-fun') and len(node[2]) == 0
                and get_sort(node[1]) is not None
                and is_bv_sort(get_sort(node[1]))
                and is_defined_fun(node[1]) and is_indexed_operator_app(
                    get_defined_fun(node[1]), 'zero_extend', 1)
                and is_defined_fun(node[-1][-1]) and is_indexed_operator_app(
                    get_defined_fun(node[-1][-1]), 'zero_extend', 1))

    def mutations(self, node):
        name = node[1]
        nsort = node[3]
        zext = int(get_defined_fun(name)[0][-1].data)
        deffun_name = node[-1][-1]
        deffun_body = get_defined_fun(deffun_name)
        deffun_zext = int(deffun_body[0][-1].data)
        decfun_name = deffun_body[-1]
        return [
            Simplification(
                {
                    node.id:
                    Node('define-fun', name, (), nsort,
                         (('_', 'zero_extend', zext + deffun_zext),
                          decfun_name))
                }, [])
        ]

    def __str__(self):
        return 'merge previous bit-width reductions'


def get_mutators():
    """Return mapping from mutator class names to the name of their config
    options."""
    return {
        'BVNormalizeConstants': 'bv-norm-constants',
        'BVConcatToZeroExtend': 'bv-zero-concat',
        'BVDoubleNegation': 'bv-double-negation',
        'BVElimBVComp': 'bv-elim-bvcomp',
        'BVEvalExtend': 'bv-eval-extend',
        'BVExtractConstants': 'bv-extract-constants',
        'BVExtractZeroExtend': 'bv-extract-zeroextend',
        'BVMergeReducedBW': 'bv-merge-reduced-bw',
        'BVIteToBVComp': 'bv-ite-to-bvcomp',
        'BVReflexiveNand': 'bv-reflexive-nand',
        'BVSimplifyConstants': 'bv-simp-constants',
        'BVTransformToBool': 'bv-to-bool',
        'BVZeroExtendPredicate': 'bv-zeroextend-pred',
        'BVReduceBW': 'bv-reduce-bitwidth',
    }


def is_relevant(node):
    """Checks whether this theory might be relevant for this node."""
    if node.has_ident():
        if node.get_ident() in ['declare-const']:
            if nodes.contains(node[2], is_bv_sort):
                return True
        elif node.get_ident() in ['declare-fun', 'define-fun', 'define-sort']:
            if nodes.contains(node[3], is_bv_sort):
                return True
    return False



#
# ddSMT: A delta debugger for SMT benchmarks in SMT-Lib v2 format.
#
# This file is part of ddSMT.
#
# Copyright (C) 2013-2021 by the authors listed in AUTHORS file.
#
# ddSMT is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ddSMT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ddSMT.  If not, see <https://www.gnu.org/licenses/>.

from . import nodes
from .smtlib import *
from .mutator_utils import Simplification


class SeqNthUnit:
    """Simplify ``(seq.nth (seq.unit t) 0)`` to ``t``."""
    def filter(self, node):
        return is_operator_app(
            node, 'seq.nth') and len(node) > 1 and is_operator_app(
                node[1], 'seq.unit')

    def mutations(self, node):
        yield Simplification({node.id: node[1][1]}, [])

    def __str__(self):
        return 'remove nth of unit seq'


class StringSimplifyConstant:
    """Replace a string constant by a shorter constant."""
    def __fix_escape_sequences(self, s, end):
        id = s.rfind('\\', end - 8, end)
        if id != -1:
            return id
        return end

    def filter(self, node):
        return is_string_const(node) and node != '""'

    def mutations(self, node):
        yield Simplification({node.id: Node('""')}, [])
        content = node[1:-1]
        for sec in nodes.binary_search(len(content)):
            start = self.__fix_escape_sequences(content, sec[0])
            yield Simplification(
                {node.id: Node(f'"{content[:start]}{content[sec[1]:]}"')}, [])
        yield Simplification({node.id: Node(f'"{content[1:]}"')}, [])
        yield Simplification({node.id: Node(f'"{content[:-1]}"')}, [])

    def global_mutations(self, node, input_):
        for simp in self.mutations(node):
            yield Simplification({node: simp.substs[node.id]}, [])

    def __str__(self):
        return 'simplify string constant'


class StringReplaceAll:
    """Replace ``str.replace_all`` by ``str.replace``."""
    def filter(self, node):
        return node.has_ident() and node.get_ident() == 'str.replace_all'

    def mutations(self, node):
        return [Simplification({node.id: Node('str.replace', *node[1:])}, [])]

    def __str__(self):
        return 'eliminate str.replace_all'


class StringIndexOfNotFound:
    """Replace ``str.indexof`` by special value ``(- 1)`` indicating that the
    substring was not found."""
    def filter(self, node):
        return node.has_ident() and node.get_ident() == 'str.indexof'

    def mutations(self, node):
        return [Simplification({node.id: Node('-', '1')}, [])]

    def __str__(self):
        return 'eliminate str.indexof'


class StringContainsToConcat:
    """Replace ``str.contains`` by concatenation."""
    def filter(self, node):
        return node.has_ident() and node.get_ident() == 'str.contains'

    def global_mutations(self, node, input_):
        var = node[1]
        k1 = f'{var}_prefix'
        k2 = f'{var}_suffix'
        vars = [
            Node('declare-const', k1, 'String'),
            Node('declare-const', k2, 'String'),
        ]
        eq = Node('=', var, ('str.++', k1, node[2], k2))
        return [Simplification({node: eq}, vars)]

    def __str__(self):
        return 'eliminate str.contains by ++'


def get_mutators():
    """Returns a mapping from mutator class names to the name of their config
    options."""
    return {
        'SeqNthUnit': 'seq-nth-unit',
        'StringContainsToConcat': 'str-contains-to-concat',
        'StringIndexOfNotFound': 'str-index-not-found',
        'StringReplaceAll': 'str-replace-all',
        'StringSimplifyConstant': 'str-constants',
    }


def is_relevant(node):
    """Checks whether this theory might be relevant for this node."""
    if node.has_ident():
        if node.get_ident() in ['declare-const']:
            if nodes.contains(node[2], lambda t: t == 'String'):
                return True
            if nodes.contains(node[2], is_seq_type):
                return True
        elif node.get_ident() in ['declare-fun', 'define-fun', 'define-sort']:
            if nodes.contains(node[3], lambda t: t == 'String'):
                return True
            if nodes.contains(node[3], is_seq_type):
                return True
    return False



from .. import mutator_utils


def check_mutations(mut, node, expected):
    assert hasattr(mut, 'mutations')

    mutations = list(mut.mutations(node))
    assert len(mutations) == len(expected)

    for res, exp in zip(mutations, expected):
        simp = mutator_utils.apply_simp(node, res)
        if simp != exp:
            return False

    return True


def check_global_mutations(mut, node, exprs, expected):
    assert hasattr(mut, 'global_mutations')

    mutations = list(mut.global_mutations(node, exprs))
    assert len(mutations) == len(expected)

    for res, exp in zip(mutations, expected):
        simp = mutator_utils.apply_simp(exprs, res)
        if simp != exp:
            return False

    return True



from ..nodes import Node
from .. import mutators_strings
from .utils import *
from ..mutator_utils import *


def test_strings_get_mutators():
    d = mutators_strings.get_mutators()
    assert isinstance(d, dict)
    assert len(d) == 5


def test_strings_is_relevant():
    assert mutators_strings.is_relevant(Node('declare-const', 'x', 'String'))
    assert mutators_strings.is_relevant(Node('declare-fun', 'x', (), 'String'))
    assert not mutators_strings.is_relevant(Node('declare-const', 'x', 'Real'))
    assert not mutators_strings.is_relevant(
        Node('declare-fun', 'x', (), 'Real'))
    assert not mutators_strings.is_relevant(Node())
    assert not mutators_strings.is_relevant(Node('assert', 'x'))


def test_strings_seqnthunit():
    expr = Node('seq.nth', ('seq.unit', 'x'), 'y')
    target = Node('x')
    m = mutators_strings.SeqNthUnit()
    assert isinstance(str(m), str)
    assert m.filter(expr)
    assert check_mutations(m, expr, [target])


def test_strings_simplify_constant():
    expr = Node('"abc123"')
    gexpr = Node('define-fun', 's', (), 'String', '"abc123"')
    m = mutators_strings.StringSimplifyConstant()
    assert isinstance(str(m), str)
    assert m.filter(expr)
    assert check_mutations(m, expr, [
        Node('""'),
        Node('"abc"'),
        Node('"123"'),
        Node('"bc123"'),
        Node('"abc12"')
    ])

    assert check_global_mutations(m, expr, gexpr, [
        apply_simp(gexpr, Simplification({Node('"abc123"'): Node('""')}, [])),
        apply_simp(gexpr, Simplification({Node('"abc123"'): Node('"abc"')},
                                         [])),
        apply_simp(gexpr, Simplification({Node('"abc123"'): Node('"123"')},
                                         [])),
        apply_simp(gexpr,
                   Simplification({Node('"abc123"'): Node('"bc123"')}, [])),
        apply_simp(gexpr,
                   Simplification({Node('"abc123"'): Node('"abc12"')}, []))
    ])

    assert check_mutations(m, Node('"abcdefg\\uab"'), [
        Node('""'),
        Node('"abcde"'),
        Node('"fg\\uab"'),
        Node('"abcdefg"'),
        Node('"abcdeuab"'),
        Node('"abfg\\uab"'),
        Node('"cdefg\\uab"'),
        Node('"bcdefg\\uab"'),
        Node('"abcdefg\\ua"'),
    ])


def test_strings_replace_all():
    expr = Node('str.replace_all', 'x', 'from', 'to')
    target = Node('str.replace', 'x', 'from', 'to')
    m = mutators_strings.StringReplaceAll()
    assert isinstance(str(m), str)
    assert m.filter(expr)
    assert check_mutations(m, expr, [target])


def test_strings_indexof_not_found():
    expr = Node('str.indexof', 'x', '0', '10')
    target = Node('-', '1')
    m = mutators_strings.StringIndexOfNotFound()
    assert isinstance(str(m), str)
    assert m.filter(expr)
    assert check_mutations(m, expr, [target])


def test_strings_contains_to_concat():
    expr = Node('str.contains', 'x', 'y')
    exprs = [
        Node('set-logic', 'QF_S'),
        Node('declare-const', 'x', 'String'),
        Node('declare-fun', 'y', (), 'String'),
        Node('assert', expr),
    ]
    target = [
        Node('set-logic', 'QF_S'),
        Node('declare-const', 'x_prefix', 'String'),
        Node('declare-const', 'x_suffix', 'String'),
        Node('declare-const', 'x', 'String'),
        Node('declare-fun', 'y', (), 'String'),
        Node('assert', ('=', 'x', ('str.++', 'x_prefix', 'y', 'x_suffix'))),
    ]
    m = mutators_strings.StringContainsToConcat()
    assert isinstance(str(m), str)
    assert m.filter(expr)
    assert check_global_mutations(m, expr, exprs, [target])



from ..nodes import Node
from .. import mutators_bv
from .. import smtlib
from .utils import *


def test_bv_get_mutators():
    d = mutators_bv.get_mutators()
    assert isinstance(d, dict)
    assert len(d) == 14


def test_bv_is_relevant():
    assert mutators_bv.is_relevant(
        Node('declare-const', 'x', ('_', 'BitVec', '17')))
    assert mutators_bv.is_relevant(
        Node('declare-fun', 'x', (), ('_', 'BitVec', '5')))
    assert mutators_bv.is_relevant(
        Node('define-sort', 'x', (), ('_', 'BitVec', '5')))
    assert not mutators_bv.is_relevant(Node('bvadd', 'x', 'y'))
    assert not mutators_bv.is_relevant(Node('declare-const', 'x', 'Int'))
    assert not mutators_bv.is_relevant(Node('declare-fun', 'x', (), 'Bool'))
    assert not mutators_bv.is_relevant(Node('x'))


def test_normalize_constants():
    m = mutators_bv.BVNormalizeConstants()
    assert isinstance(str(m), str)
    assert not m.filter(Node(3))
    assert not m.filter(Node('define-fun', 'f', (), '#b0010'))
    assert not m.filter(Node('_', 'bv4', 4))
    assert not m.filter(Node('#b2010'))
    assert not m.filter(Node('#xz010'))
    assert m.filter(Node('#b0010'))
    assert m.filter(Node('#xa1f'))
    assert check_mutations(m, Node('#b0'), [('_', 'bv0', 1)])
    assert check_mutations(m, Node('#b1'), [('_', 'bv1', 1)])
    assert check_mutations(m, Node('#b000'), [('_', 'bv0', 3)])
    assert check_mutations(m, Node('#b001'), [('_', 'bv1', 3)])
    assert check_mutations(m, Node('#b010'), [('_', 'bv2', 3)])
    assert check_mutations(m, Node('#b011'), [('_', 'bv3', 3)])
    assert check_mutations(m, Node('#b100'), [('_', 'bv4', 3)])
    assert check_mutations(m, Node('#b101'), [('_', 'bv5', 3)])
    assert check_mutations(m, Node('#b110'), [('_', 'bv6', 3)])
    assert check_mutations(m, Node('#b111'), [('_', 'bv7', 3)])
    assert check_mutations(m, Node('#x11'), [('_', 'bv17', 8)])
    assert check_mutations(m, Node('#x011'), [('_', 'bv17', 12)])
    assert check_mutations(m, Node('#xa11'), [('_', 'bv2577', 12)])


def test_bv_concat_to_zero_extend():
    smtlib.reset_information()
    m = mutators_bv.BVConcatToZeroExtend()
    assert isinstance(str(m), str)
    r = Node('r')
    x = Node('x')
    const = Node('declare-const', r, 'Real')
    val8_0 = '#b00000000'
    val8_1 = '#b01011100'
    conc16_0 = Node('concat', x, val8_0)
    conc16_1 = Node('concat', x, val8_1)
    conc16_2 = Node('concat', x, x)
    conc16_3 = Node('concat', x, r)
    conc16_4 = Node('concat', val8_0, x)
    exprs = [conc16_0, conc16_1, conc16_2, conc16_3, conc16_4]
    smtlib.collect_information(exprs)
    assert not m.filter(const)
    assert not m.filter(conc16_0)
    assert not m.filter(conc16_1)
    assert not m.filter(conc16_2)
    assert not m.filter(conc16_3)
    assert m.filter(conc16_4)
    assert check_mutations(m, conc16_4, [Node(('_', 'zero_extend', 8), 'x')])


def test_bv_double_negation():
    m = mutators_bv.BVDoubleNegation()
    assert isinstance(str(m), str)
    x = Node('x')
    bvnot = Node('bvnot', Node('bvnot', x))
    bvneg = Node('bvneg', Node('bvneg', x))
    assert not m.filter(Node('bvadd', x, x))
    assert m.filter(bvnot)
    assert m.filter(bvneg)
    assert check_mutations(m, bvnot, [x])
    assert check_mutations(m, bvneg, [x])


def test_bv_elim_bvcomp():
    m = mutators_bv.BVElimBVComp()
    assert isinstance(str(m), str)
    x = Node('x')
    const = Node('#011')
    const0 = Node('#b0')
    const1 = Node('#b1')
    bvcomp = Node('bvcomp', x, x)
    eq0_1 = Node('=', const0, bvcomp)
    eq1_1 = Node('=', const0, bvcomp, x)
    eq2_1 = Node('=', const0, x, bvcomp, x)
    eq0_2 = Node('=', const1, bvcomp)
    eq1_2 = Node('=', const1, bvcomp, x)
    eq2_2 = Node('=', const1, x, bvcomp, x)
    eq_x = Node('=', x, x)
    ne_x = Node('not', eq_x)
    assert not m.filter(x)
    assert not m.filter(Node('=', x, x))
    assert not m.filter(Node('=', const0, x, x))
    assert not m.filter(Node('=', const, bvcomp))
    assert m.filter(eq0_1)
    assert m.filter(eq1_1)
    assert m.filter(eq2_1)
    assert m.filter(eq0_2)
    assert m.filter(eq1_2)
    assert m.filter(eq2_2)
    assert check_mutations(m, eq0_1, [ne_x])
    assert check_mutations(m, eq1_1, [Node('and', ne_x, ('=', const0, x))])
    assert check_mutations(
        m, eq2_1, [Node('and', ('=', const0, x), ne_x, ('=', const0, x))])
    assert check_mutations(m, eq0_2, [Node('=', x, x)])
    assert check_mutations(m, eq1_2, [Node('and', eq_x, ('=', const1, x))])
    assert check_mutations(
        m, eq2_2, [Node('and', ('=', const1, x), eq_x, ('=', const1, x))])


def test_bv_eval_extend():
    m = mutators_bv.BVEvalExtend()
    assert isinstance(str(m), str)
    x = Node('x')
    const0_1 = Node('#b011')
    const0_2 = Node('_', 'bv3', 3)
    const1_1 = Node('#b101')
    const1_2 = Node('_', 'bv5', 3)
    zext0_1 = Node(('_', 'zero_extend', 2), const0_1)
    zext0_2 = Node(('_', 'zero_extend', 2), const0_2)
    zext1_1 = Node(('_', 'zero_extend', 2), const1_1)
    zext1_2 = Node(('_', 'zero_extend', 2), const1_2)
    sext0_1 = Node(('_', 'sign_extend', 2), const0_1)
    sext0_2 = Node(('_', 'sign_extend', 2), const0_2)
    sext1_1 = Node(('_', 'sign_extend', 2), const1_1)
    sext1_2 = Node(('_', 'sign_extend', 2), const1_2)
    assert not m.filter(Node('bvand', x, x))
    assert not m.filter(Node(('_', 'zero_extend', 2), x))
    assert not m.filter(Node(('zero_extend', 2), const0_1))
    assert m.filter(zext0_1)
    assert m.filter(zext0_2)
    assert m.filter(zext1_1)
    assert m.filter(zext1_2)
    assert m.filter(sext0_1)
    assert m.filter(sext0_2)
    assert m.filter(sext1_1)
    assert m.filter(sext1_2)
    assert check_mutations(m, zext0_1, [Node('_', 'bv3', 5)])
    assert check_mutations(m, zext0_2, [Node('_', 'bv3', 5)])
    assert check_mutations(m, zext1_1, [Node('_', 'bv5', 5)])
    assert check_mutations(m, zext1_2, [Node('_', 'bv5', 5)])
    assert check_mutations(m, sext0_1, [Node('_', 'bv3', 5)])
    assert check_mutations(m, sext0_2, [Node('_', 'bv3', 5)])
    assert check_mutations(m, sext1_1, [Node('#b11101')])
    assert check_mutations(m, sext1_2, [Node('#b11101')])


def test_bv_extract_constants():
    m = mutators_bv.BVExtractConstants()
    assert isinstance(str(m), str)
    const0 = Node('#b011')
    const1 = Node('_', 'bv3', 3)
    const2 = Node('_', 'bv7', 3)
    ext0_1 = Node(('_', 'extract', 0, 0), const0)
    ext0_2 = Node(('_', 'extract', 1, 0), const0)
    ext0_3 = Node(('_', 'extract', 2, 0), const0)
    ext1_1 = Node(('_', 'extract', 2, 0), const1)
    ext1_2 = Node(('_', 'extract', 2, 1), const1)
    ext1_3 = Node(('_', 'extract', 2, 2), const1)
    ext2_1 = Node(('_', 'extract', 1, 1), const2)
    ext2_2 = Node(('_', 'extract', 2, 1), const2)
    ext2_3 = Node(('_', 'extract', 2, 2), const2)
    assert not m.filter(Node(('_', 'extract', 0, 0), ('bvadd', 'x', 'y')))
    assert m.filter(ext0_1)
    assert m.filter(ext0_2)
    assert m.filter(ext0_3)
    assert m.filter(ext1_1)
    assert m.filter(ext1_2)
    assert m.filter(ext1_3)
    assert check_mutations(m, ext0_1, [Node('#b1')])
    assert check_mutations(m, ext0_2, [Node('#b11')])
    assert check_mutations(m, ext0_3, [Node('#b011')])
    assert check_mutations(m, ext1_1, [Node('#b011')])
    assert check_mutations(m, ext1_2, [Node('#b01')])
    assert check_mutations(m, ext1_3, [Node('#b0')])
    assert check_mutations(m, ext2_1, [Node('#b1')])
    assert check_mutations(m, ext2_2, [Node('#b11')])
    assert check_mutations(m, ext2_3, [Node('#b1')])


def test_bv_extract_zero_extend():
    smtlib.reset_information()
    m = mutators_bv.BVExtractZeroExtend()
    assert isinstance(str(m), str)
    x = Node('x')
    ext1 = Node(('_', 'extract', 7, 5), (('_', 'zero_extend', 4), x))
    ext2 = Node(('_', 'extract', 1, 0), (('_', 'zero_extend', 4), x))
    ext3 = Node(('_', 'extract', 5, 3), (('_', 'zero_extend', 4), x))
    exprs = [Node('declare-const', x, ('_', 'BitVec', 4)), ext1, ext2, ext3]
    assert not m.filter(Node(('_', 'extract', 7, 5), x))
    assert m.filter(ext1)
    assert m.filter(ext2)
    assert m.filter(ext3)
    assert check_mutations(m, ext1, [])
    smtlib.collect_information(exprs)
    assert check_mutations(m, ext1, [Node('_', 'bv0', 3)])
    assert check_mutations(m, ext2, [Node(('_', 'extract', 1, 0), x)])
    assert check_mutations(
        m, ext3, [Node(('_', 'zero_extend', 2), (('_', 'extract', 3, 3), x))])


def test_bv_ite_to_bvcomp():
    smtlib.reset_information()
    m = mutators_bv.BVIteToBVComp()
    assert isinstance(str(m), str)
    assert not m.filter(Node('x'))
    assert not m.filter(Node('=', 'x', 'y'))
    assert not m.filter(Node('ite', 'c', 'x', 'y'))
    assert not m.filter(Node('ite', ('=', 'z', 'z'), '#b1', '#b0'))
    assert not m.filter(Node('ite', ('=', 'x', 'y'), '#b1', '#b0'))
    declx = Node('declare-const', 'x', ('_', 'BitVec', 4))
    decly = Node('declare-const', 'y', ('_', 'BitVec', 4))
    declz = Node('declare-const', 'z', 'Int')
    ite = Node('ite', ('=', 'x', 'y'), '#b1', '#b0')
    exprs = [declx, decly, declz, ite]
    smtlib.collect_information(exprs)
    assert not m.filter(Node('ite', ('=', 'z', 'z'), '#b1', '#b0'))
    assert not m.filter(Node('ite', ('=', 'x', 'y'), '#b1', '#b1'))
    assert not m.filter(Node('ite', ('=', 'x', 'y'), '#b11', '#b01'))
    assert m.filter(ite)
    assert check_mutations(m, ite, [Node('bvcomp', 'x', 'y')])


def test_reflexive_nand():
    m = mutators_bv.BVReflexiveNand()
    assert isinstance(str(m), str)
    assert not m.filter(Node('x'))
    assert not m.filter(Node('bvor', 'x', 'y'))
    assert not m.filter(Node('bvnand', 'x', 'y'))
    assert m.filter(Node('bvnand', 'x', 'x'))
    assert check_mutations(m, Node('bvnand', 'x', 'x'), [Node('bvnot', 'x')])


def test_simplify_constants():
    m = mutators_bv.BVSimplifyConstants()
    assert isinstance(str(m), str)
    assert not m.filter(Node('x'))
    assert not m.filter(Node('declare-const', 'x', ('_', 'BitVec', 4)))
    assert not m.filter(Node('0'))
    assert not m.filter(Node('1'))
    assert not m.filter(Node('#x0'))
    assert not m.filter(Node('#x1'))
    assert not m.filter(Node('#b0'))
    assert not m.filter(Node('#b1'))
    assert not m.filter(Node('_', 'bv0', 8))
    assert not m.filter(Node('_', 'bv1', 8))
    assert m.filter(Node('#b10'))
    assert m.filter(Node('#b11'))
    assert m.filter(Node('_', 'bv4', 3))
    assert check_mutations(m, Node('#b10'), [Node('#b00'), Node('#b01')])
    bvconst = Node('#b11001010')
    assert check_mutations(m, bvconst, [
        Node('#b00000000'),
        Node('#b00000001'),
        Node('#b00000110'),
        Node('#b00011001'),
        Node('#b01100101'),
    ])
    assert check_global_mutations(
        m, bvconst, [Node('assert', ('=', bvconst, 'x'))], [
            [Node('assert', ('=', Node('#b00000000'), 'x'))],
            [Node('assert', ('=', Node('#b00000001'), 'x'))],
            [Node('assert', ('=', Node('#b00000110'), 'x'))],
            [Node('assert', ('=', Node('#b00011001'), 'x'))],
            [Node('assert', ('=', Node('#b01100101'), 'x'))],
        ])


def test_transform_to_bool():
    m = mutators_bv.BVTransformToBool()
    assert isinstance(str(m), str)
    assert not m.filter(Node('x'))
    assert not m.filter(Node('=', 'x', 'y'))
    assert not m.filter(Node('=', '#b11', 'x'))
    assert not m.filter(Node('=', 'x', '#b1'))
    assert not m.filter(Node('=', '#b1', 'x'))
    assert not m.filter(Node('=', '#b1', ('bvand', 'x', 'y'), 'z'))
    assert not m.filter(Node('=', '#b1', ('_', 'bv0', '1')))
    assert m.filter(Node('=', '#b1', ('bvand', 'x', 'y')))
    assert m.filter(Node('=', ('bvand', 'x', 'y'), '#b1'))
    assert m.filter(Node('=', '#b1', ('bvand', 'x', 'y', 'z')))
    assert check_mutations(
        m, Node('=', '#b1', ('bvand', 'x', 'y', 'z')),
        [Node('and', ('=', '#b1', 'x'), ('=', '#b1', 'y'), ('=', '#b1', 'z'))])
    assert check_mutations(
        m, Node('=', ('bvand', 'x', 'y', 'z'), '#b1'),
        [Node('and', ('=', '#b1', 'x'), ('=', '#b1', 'y'), ('=', '#b1', 'z'))])
    assert check_mutations(
        m, Node('=', '#b0', ('bvor', 'x', 'y', 'z')),
        [Node('or', ('=', '#b0', 'x'), ('=', '#b0', 'y'), ('=', '#b0', 'z'))])
    assert check_mutations(
        m, Node('=', ('bvor', 'x', 'y', 'z'), '#b0'),
        [Node('or', ('=', '#b0', 'x'), ('=', '#b0', 'y'), ('=', '#b0', 'z'))])
    assert check_mutations(
        m, Node('=', '#b1', ('bvxor', 'x', 'y', 'z')),
        [Node('xor', ('=', '#b1', 'x'), ('=', '#b1', 'y'), ('=', '#b1', 'z'))])
    assert check_mutations(
        m, Node('=', ('bvxor', 'x', 'y', 'z'), '#b0'),
        [Node('xor', ('=', '#b0', 'x'), ('=', '#b0', 'y'), ('=', '#b0', 'z'))])


def test_zero_extend_predicate():
    m = mutators_bv.BVZeroExtendPredicate()
    assert isinstance(str(m), str)
    assert not m.filter(Node('x'))
    assert not m.filter(Node('=', 'x', 'y'))
    assert not m.filter(Node('=', (('_', 'zero_extend', 2), 'x'), 'y'))
    assert not m.filter(Node('=', 'x', (('_', 'zero_extend', 2), 'y')))
    assert not m.filter(
        Node('bvadd', (('_', 'zero_extend', 2), 'x'),
             (('_', 'zero_extend', 2), 'y')))
    assert not m.filter(
        Node('=', (('_', 'zero_extend', 2), 'x'),
             (('_', 'zero_extend', 2), 'y'), (('_', 'zero_extend', 2), 'z')))
    assert m.filter(
        Node('=', (('_', 'zero_extend', 2), 'x'),
             (('_', 'zero_extend', 2), 'y')))
    assert m.filter(
        Node('distinct', (('_', 'zero_extend', 2), 'x'),
             (('_', 'zero_extend', 2), 'y')))
    assert m.filter(
        Node('bvult', (('_', 'zero_extend', 2), 'x'),
             (('_', 'zero_extend', 2), 'y')))
    assert m.filter(
        Node('bvule', (('_', 'zero_extend', 2), 'x'),
             (('_', 'zero_extend', 2), 'y')))
    assert m.filter(
        Node('bvugt', (('_', 'zero_extend', 2), 'x'),
             (('_', 'zero_extend', 2), 'y')))
    assert m.filter(
        Node('bvuge', (('_', 'zero_extend', 2), 'x'),
             (('_', 'zero_extend', 2), 'y')))
    assert m.filter(
        Node('bvslt', (('_', 'zero_extend', 2), 'x'),
             (('_', 'zero_extend', 2), 'y')))
    assert m.filter(
        Node('bvsle', (('_', 'zero_extend', 2), 'x'),
             (('_', 'zero_extend', 2), 'y')))
    assert m.filter(
        Node('bvsgt', (('_', 'zero_extend', 2), 'x'),
             (('_', 'zero_extend', 2), 'y')))
    assert m.filter(
        Node('bvsge', (('_', 'zero_extend', 2), 'x'),
             (('_', 'zero_extend', 2), 'y')))
    assert check_mutations(
        m,
        Node('=', (('_', 'zero_extend', 2), 'x'),
             (('_', 'zero_extend', 2), 'y')), [Node('=', 'x', 'y')])
    assert check_mutations(
        m,
        Node('=', (('_', 'zero_extend', 3), 'x'),
             (('_', 'zero_extend', 2), 'y')),
        [Node('=', (('_', 'zero_extend', 1), 'x'), 'y')])
    assert check_mutations(
        m,
        Node('=', (('_', 'zero_extend', 3), 'x'),
             (('_', 'zero_extend', 5), 'y')),
        [Node('=', 'x', (('_', 'zero_extend', 2), 'y'))])


def test_bv_reduce_bw():
    smtlib.reset_information()
    m = mutators_bv.BVReduceBW()
    assert isinstance(str(m), str)
    x = Node('x')
    y = Node('y')
    z = Node('z')
    r = Node('r')
    s = Node('s')
    a = Node('a')
    b = Node('b')
    c = Node('c')
    bvsort8 = Node('_', 'BitVec', 8)
    bvsort9 = Node('_', 'BitVec', 9)
    assert not m.filter(Node('bvadd', x, y))
    assert not m.filter(Node('declare-const', x, 'Int'))
    assert not m.filter(Node('declare-fun', y, ('Int', ), bvsort8))
    declx = Node('declare-const', x, bvsort8)
    decly = Node('declare-fun', y, (), bvsort8)
    assert not m.filter(declx)
    assert not m.filter(decly)
    exprs0 = [declx, decly]
    smtlib.collect_information(exprs0)
    assert m.filter(declx)
    assert m.filter(decly)
    declz = Node('declare-const', z, bvsort9)
    _decls1 = [
        Node('declare-fun', r, (), 'Real'),
        Node('declare-const', s, 'String'),
        Node('declare-const', a, 'Float64'),
        Node('declare-const', b, ('_', 'FloatingPoint', 11, 53)),
        Node('declare-fun', c, (), ('_', 'FloatingPoint', 5, 11)),
    ]
    _asserts1 = [
        Node('assert', ('=', ('bvadd', x, y), (('_', 'extract', 24, 17), z))),
        Node('assert', ('<', ('+', r, r), 4)),
        Node('assert', ('distinct', ('str.++', s, s), "aaa")),
    ]
    exprs1 = [
        declx,
        decly,
        declz,
        *_decls1,
        *_asserts1,
    ]
    x8 = [
        Node('declare-const', '_x', ('_', 'BitVec', 1)),
        Node('declare-const', '_x', ('_', 'BitVec', 2)),
        Node('declare-const', '_x', ('_', 'BitVec', 4)),
        Node('declare-const', '_x', ('_', 'BitVec', 7)),
    ]
    y8 = [
        Node('declare-const', '_y', ('_', 'BitVec', 1)),
        Node('declare-const', '_y', ('_', 'BitVec', 2)),
        Node('declare-const', '_y', ('_', 'BitVec', 4)),
        Node('declare-const', '_y', ('_', 'BitVec', 7)),
    ]
    z9 = [
        Node('declare-const', '_z', ('_', 'BitVec', 1)),
        Node('declare-const', '_z', ('_', 'BitVec', 2)),
        Node('declare-const', '_z', ('_', 'BitVec', 4)),
        Node('declare-const', '_z', ('_', 'BitVec', 8)),
    ]
    zext_x8 = [
        Node(('_', 'zero_extend', 7), '_x'),
        Node(('_', 'zero_extend', 6), '_x'),
        Node(('_', 'zero_extend', 4), '_x'),
        Node(('_', 'zero_extend', 1), '_x'),
    ]
    zext_y8 = [
        Node(('_', 'zero_extend', 7), '_y'),
        Node(('_', 'zero_extend', 6), '_y'),
        Node(('_', 'zero_extend', 4), '_y'),
        Node(('_', 'zero_extend', 1), '_y'),
    ]
    zext_z9 = [
        Node(('_', 'zero_extend', 8), '_z'),
        Node(('_', 'zero_extend', 7), '_z'),
        Node(('_', 'zero_extend', 5), '_z'),
        Node(('_', 'zero_extend', 1), '_z'),
    ]
    smtlib.collect_information(exprs1)
    assert check_global_mutations(m, declx, exprs1, [[
        x8[i],
        Node('define-fun', x, (), bvsort8, zext_x8[i]),
        decly,
        declz,
        *_decls1,
        *_asserts1,
    ] for i in range(0, 4)])
    assert check_global_mutations(m, decly, exprs1, [[
        y8[i],
        declx,
        Node('define-fun', y, (), bvsort8, zext_y8[i]),
        declz,
        *_decls1,
        *_asserts1,
    ] for i in range(0, 4)])
    assert check_global_mutations(m, declz, exprs1, [[
        z9[i],
        declx,
        decly,
        Node('define-fun', z, (), bvsort9, zext_z9[i]),
        *_decls1,
        *_asserts1,
    ] for i in range(0, 4)])

    # corner case: avoid reducing by 0 bits
    decl = Node('declare-const', 'x', ('_', 'BitVec', 2))
    smtlib.collect_information([decl])
    assert m.filter(decl)
    assert check_global_mutations(m, decl, [decl], [[
        Node('declare-const', '_x', ('_', 'BitVec', 1)),
        Node('define-fun', 'x', (), ('_', 'BitVec', 2),
             (('_', 'zero_extend', '1'), '_x')),
    ]])

    decl = Node('declare-const', 'x', ('_', 'BitVec', 1))
    smtlib.collect_information([decl])
    assert m.filter(decl)
    assert check_global_mutations(m, decl, [decl], [])


def test_bv_merge_reduced_bw():
    smtlib.reset_information()
    m = mutators_bv.BVMergeReducedBW()
    assert isinstance(str(m), str)
    bvsort1 = Node('_', 'BitVec', 1)
    bvsort4 = Node('_', 'BitVec', 4)
    bvsort8 = Node('_', 'BitVec', 8)

    assert not m.filter(Node('bvadd', 'x', 'y'))
    assert not m.filter(Node('declare-const', 'x', 'Int'))
    assert not m.filter(Node('declare-fun', 'y', (), bvsort8))
    assert not m.filter(
        Node('define-fun', 'y', ('Int', ), bvsort8,
             (('_', 'zero_extend', 2), '_y')))

    dec_x__ = Node('declare-const', '__x', bvsort1)
    def_x_ = Node('define-fun', '_x', (), bvsort4,
                  (('_', 'zero_extend', 3), '__x'))
    def_x = Node('define-fun', 'x', (), bvsort8,
                 (('_', 'zero_extend', 4), '_x'))
    def_y = Node('define-fun', 'y', (), bvsort8,
                 (('_', 'zero_extend', 2), '__x'))
    assert not m.filter(def_y)
    assert not m.filter(def_x)

    exprs = [dec_x__, def_x_, def_x, def_y]
    smtlib.collect_information(exprs)
    assert not m.filter(dec_x__)
    assert not m.filter(def_x_)
    assert not m.filter(def_y)
    assert m.filter(exprs[2])
    assert m.filter(def_x)
    assert check_mutations(m, def_x, [
        Node('define-fun', 'x', (), bvsort8, (('_', 'zero_extend', 7), '__x'))
    ])


def test_bv_reduce_bw_usecase():
    from .. import mutators_core
    from .. import mutators_smtlib
    from .. import mutator_utils
    from .. import nodeio
    from .. import nodes

    exprs = [
        Node('declare-const', 'x', ('_', 'BitVec', '32')),
        Node('assert', ('=', 'x', ('concat', (('_', 'extract', 15, 0), 'x'),
                                   (('_', 'extract', 31, 16), 'x'))))
    ]
    print('\n'.join([nodeio.__write_smtlib_str(expr) for expr in exprs]))
    smtlib.collect_information(exprs)

    mrbw = mutators_bv.BVReduceBW()
    assert mrbw.filter(exprs[0])
    mut = list(mrbw.global_mutations(exprs[0], exprs))[2]
    assert mut.fresh_vars[0][2][2] == '16'
    exprs = mutator_utils.apply_simp(exprs, mut)
    print('\n'.join([nodeio.__write_smtlib_str(expr) for expr in exprs]))
    smtlib.collect_information(exprs)

    midf = mutators_smtlib.InlineDefinedFuns()
    for n in nodes.dfs(exprs):
        if n == 'x':
            assert midf.filter(n)
            repllist = midf.mutations(n)
            if repllist:
                exprs = mutator_utils.apply_simp(exprs, repllist[0])

    refexprs = [
        Node('declare-const', '_x', ('_', 'BitVec', '16')),
        Node('define-fun', 'x', (), ('_', 'BitVec', '32'),
             (('_', 'zero_extend', '16'), '_x')),
        Node('assert', ('=', (('_', 'zero_extend', '16'), '_x'),
                        ('concat', (('_', 'extract', 15, 0),
                                    (('_', 'zero_extend', '16'), '_x')),
                         (('_', 'extract', 31, 16),
                          (('_', 'zero_extend', '16'), '_x')))))
    ]
    assert exprs == refexprs
    print('\n'.join([nodeio.__write_smtlib_str(expr) for expr in exprs]))

    en = mutators_core.EraseNode()
    assert en.filter(exprs[1])
    exprs = mutator_utils.apply_simp(exprs, en.mutations(exprs[1])[0])
    print('\n'.join([nodeio.__write_smtlib_str(expr) for expr in exprs]))

    bveze = mutators_bv.BVExtractZeroExtend()
    for n in [exprs[1][1][2][1], exprs[1][1][2][2]]:
        assert bveze.filter(n)
        exprs = mutator_utils.apply_simp(exprs, bveze.mutations(n)[0])
    print('\n'.join([nodeio.__write_smtlib_str(expr) for expr in exprs]))

    sorter = mutators_core.SortChildren()
    assert sorter.filter(exprs[1][1][2])
    exprs = mutator_utils.apply_simp(exprs,
                                     sorter.mutations(exprs[1][1][2])[0])
    print('\n'.join([nodeio.__write_smtlib_str(expr) for expr in exprs]))

    bvec = mutators_bv.BVConcatToZeroExtend()
    assert bvec.filter(exprs[1][1][2])
    exprs = mutator_utils.apply_simp(exprs, bvec.mutations(exprs[1][1][2])[0])
    print('\n'.join([nodeio.__write_smtlib_str(expr) for expr in exprs]))

    rbc = mutators_core.ReplaceByChild()
    assert rbc.filter(exprs[1][1][2][1])
    exprs = mutator_utils.apply_simp(exprs,
                                     list(rbc.mutations(exprs[1][1][2][1]))[0])
    print('\n'.join([nodeio.__write_smtlib_str(expr) for expr in exprs]))

    bvzep = mutators_bv.BVZeroExtendPredicate()
    assert bvzep.filter(exprs[1][1])
    exprs = mutator_utils.apply_simp(exprs, bvzep.mutations(exprs[1][1])[0])
    print('\n'.join([nodeio.__write_smtlib_str(expr) for expr in exprs]))



#
# ddSMT: A delta debugger for SMT benchmarks in SMT-Lib v2 format.
#
# This file is part of ddSMT.
#
# Copyright (C) 2013-2021 by the authors listed in AUTHORS file.
#
# ddSMT is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ddSMT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ddSMT.  If not, see <https://www.gnu.org/licenses/>.

import os
import shutil
import tempfile
import threading

from . import options

# The temporary directory. Automatically deleted upon termination.
__TMPDIR = None
__BINARY = None
__BINARY_CC = None
__FILEEXT = None


def init():
    """Create temporary directory, set filenames."""
    global __TMPDIR
    global __BINARY
    global __BINARY_CC
    global __FILEEXT
    __TMPDIR = tempfile.TemporaryDirectory(prefix="ddsmt-")
    __BINARY = os.path.join(__TMPDIR.name, 'binary')
    __BINARY_CC = os.path.join(__TMPDIR.name, 'binary_cc')
    __FILEEXT = os.path.splitext(options.args().infile)[1]


def copy_binaries():
    """Copy the solver binary from ``cmd`` to our temporary directory.

    If a cross check is defined, also copies this binary. After copying,
    it modifies the ``cmd`` argument to use the copied binary.
    """
    shutil.copy(options.args().cmd[0], __BINARY)
    options.args().cmd[0] = __BINARY

    if options.args().cmd_cc:
        shutil.copy(options.args().cmd_cc[0], __BINARY_CC)
        options.args().cmd_cc[0] = __BINARY_CC


def get_tmp_filename():
    """Return a filename within our temporary directory based on the pid."""
    return os.path.join(
        __TMPDIR.name,
        f'ddsmt-tmp-{os.getpid()}-{threading.get_ident()}{__FILEEXT}')


def copy_to_tmp_file(source):
    """Copy the given source file to a temporary file as returned by
    ``get_tmp_filename()``."""
    res = get_tmp_filename()
    shutil.copy(source, res.name)
    return res



#
# ddSMT: A delta debugger for SMT benchmarks in SMT-Lib v2 format.
#
# This file is part of ddSMT.
#
# Copyright (C) 2013-2021 by the authors listed in AUTHORS file.
#
# ddSMT is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ddSMT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ddSMT.  If not, see <https://www.gnu.org/licenses/>.

from . import nodes
from .smtlib import *
from .mutator_utils import Simplification


def is_arithmetic_relation(node):
    if not node.has_ident():
        return False
    return node.get_ident() in [
        '=', '<', '>', '>=', '<=', '!=', '<>', 'distinct'
    ]


class ArithmeticNegateRelation:
    """Replace a negated relation by the corresponding inverse relation.

    For example, ``(not (< a b))`` is replaced with ``(>= a b)``.
    """
    def filter(self, node):
        return is_operator_app(
            node, 'not') and len(node) > 1 and is_arithmetic_relation(node[1])

    def mutations(self, node):
        negator = {
            '=': 'distinct',
            '<': '>=',
            '>': '<=',
            '>=': '<',
            '<=': '>',
            '!=': '=',
            '<>': '=',
            'distinct': '='
        }
        assert node[1][0] in negator
        return [
            Simplification({node.id: Node(negator[node[1][0]], *node[1][1:])},
                           [])
        ]

    def __str__(self):
        return 'push negation into arithmetic relation'


class ArithmeticSimplifyConstant:
    """Replace a constant with a simpler constant.

    We consider constants that are smaller (in value) or with fewer decimal
    places as `simpler`.
    """
    def filter(self, node):
        return is_arith_const(node) and float(node.data) not in [0, 1]

    def mutations(self, node):
        f = float(node.data)
        if int(f) == f:
            i = int(f)
            yield from [
                Simplification({node.id: Node(str(i // 2))}, []),
                Simplification({node.id: Node(str(i // 10))}, [])
            ]
            return
        yield from [
            Simplification({node.id: Node(str(int(f)))}, []),
            Simplification({node.id: Node(node.data[:-1])}, [])
        ]

    def global_mutations(self, node, input_):
        yield from [
            Simplification({node: simp.substs[node.id]}, [])
            for simp in self.mutations(node)
        ]

    def __str__(self):
        return 'simplify arithmetic constant'


class ArithmeticSplitNaryRelation:
    """Split an n-ary relation into the conjunction of multiple relations.

    This assumes that the relation symbol is transitive.
    """
    def filter(self, node):
        return is_arithmetic_relation(node) and len(node) > 3

    def mutations(self, node):
        split = [(node.get_ident(), node[i], node[i + 1])
                 for i in range(1,
                                len(node) - 1)]
        return [Simplification({node.id: Node('and', *split)}, [])]

    def __str__(self):
        return 'split arithmetic n-ary relation'


class ArithmeticStrengthenRelation:
    """Replace a relation by a stronger relation.

    For example, ``>=`` is replaced by ``>``.
    """
    def filter(self, node):
        return is_arithmetic_relation(node) and node.get_ident() in [
            '<', '>', '<=', '>=', 'distinct'
        ]

    def mutations(self, node):
        strengthen = {
            '<': ['='],
            '>': ['='],
            '<=': ['<', '='],
            '>=': ['>', '='],
            'distinct': ['=']
        }
        assert node[0] in strengthen
        yield from [
            Simplification({node.id: Node(rel, *node.data[1:])}, [])
            for rel in strengthen[node[0]]
        ]

    def __str__(self):
        return 'strengthen arithmetic relation'


def get_mutators():
    """Returns a mapping from mutator class names to the name of their config
    options."""
    return {
        'ArithmeticNegateRelation': 'arith-negate-relations',
        'ArithmeticSimplifyConstant': 'arith-constants',
        'ArithmeticSplitNaryRelation': 'arith-split-nary-relations',
        'ArithmeticStrengthenRelation': 'arith-strengthen-relations',
    }


def is_relevant(node):
    """Checks whether this theory might be relevant for this node."""
    if node.has_ident():
        if node.get_ident() in ['declare-const']:
            if nodes.contains(node[2], lambda t: t in ['Int', 'Real']):
                return True
        elif node.get_ident() in ['declare-fun', 'define-fun', 'define-sort']:
            if nodes.contains(node[3], lambda t: t in ['Int', 'Real']):
                return True
    return False



from ..nodes import Node
from .. import mutators_smtlib
from .. import smtlib
from ..mutator_utils import *
from .utils import *


def test_smtlib_get_mutators():
    d = mutators_smtlib.get_mutators()
    assert isinstance(d, dict)
    assert len(d) == 10


def test_smtlib_checksat_assuming():
    m = mutators_smtlib.CheckSatAssuming()
    assert isinstance(str(m), str)
    assert not m.filter(Node(Node('check-sat')))
    assert not m.filter(Node('assert', 'true'))
    assert m.filter(Node('check-sat-assuming', 'true'))
    assert m.filter(Node('check-sat-assuming', ('xor', 'x', 'y')))

    assert check_mutations(m, Node('check-sat-assuming', 'true'),
                           [Node(Node('check-sat'))])
    assert check_mutations(m, Node('check-sat-assuming', ('xor', 'x', 'y')),
                           [Node(Node('check-sat'))])


def test_smtlib_eliminate_variable():
    m = mutators_smtlib.EliminateVariable()
    assert isinstance(str(m), str)
    # special case where 'x' occurs in the replacement
    eq = Node('=', 'x', ('*', 'x', 'y'))
    assert m.filter(eq)
    assert check_global_mutations(m, eq, eq, [])
    eq = Node('=', 'x', ('*', 'y', 'y'))
    exprs = [
        Node('declare-const', 'x', 'Real'),
        Node('define-fun', 'y', (), 'Real', ('*', 'x', 'x')),
        Node('define-fun', 'z', (), 'Real', ('*', 'x', 'y')),
        Node('assert', eq),
        Node('assert', ('>', 'x', 'y')),
    ]
    smtlib.collect_information(exprs)
    assert m.filter(eq)
    assert check_global_mutations(m, eq, exprs, [[
        Node('declare-const', 'x', 'Real'),
        Node('define-fun', 'y', (), 'Real',
             ('*', ('*', 'y', 'y'), ('*', 'y', 'y'))),
        Node('define-fun', 'z', (), 'Real', ('*', ('*', 'y', 'y'), 'y')),
        Node('assert', ('=', ('*', 'y', 'y'), ('*', 'y', 'y'))),
        Node('assert', ('>', ('*', 'y', 'y'), 'y')),
    ]])

    eq = Node('=', 'x', 'y', 'a', ('*', 'y', 'y'), ('+', 'a', 'b'))
    exprs = [Node('assert', ('>', 'x', 'y'))]
    smtlib.collect_information(exprs)
    assert m.filter(eq)
    assert check_global_mutations(m, eq, exprs, [
        [Node('assert', ('>', 'y', 'y'))],
        [Node('assert', ('>', 'a', 'y'))],
        [Node('assert', ('>', ('*', 'y', 'y'), 'y'))],
        [Node('assert', ('>', ('+', 'a', 'b'), 'y'))],
        [Node('assert', ('>', 'x', 'x'))],
        [Node('assert', ('>', 'x', 'a'))],
        [Node('assert', ('>', 'x', ('+', 'a', 'b')))],
    ])


def test_smtlib_inline_define_fun():
    smtlib.reset_information()
    m = mutators_smtlib.InlineDefinedFuns()
    assert isinstance(str(m), str)
    # special case where we should not attempt to inline into itself
    inner = Node('x')
    n = Node('define-fun', 'x', (), 'Bool', ('not', inner))
    smtlib.collect_information([n])
    assert m.filter(inner)
    assert check_mutations(m, inner, [])

    n = Node('define-fun', 'x', (), 'Bool', inner)
    smtlib.collect_information([n])
    assert m.mutations(inner) == []

    inner = Node('x')
    expr = Node(inner, 'a', 'b')
    smtlib.collect_information([
        Node('define-fun', 'x', (('y', 'Bool'), ('z', 'Bool')), 'Bool',
             ('and', 'y', 'z'))
    ])
    assert m.filter(expr)
    assert check_mutations(m, expr, [Node('and', 'a', 'b')])
    smtlib.reset_information()


def test_smtlib_introduce_fresh_variable():
    smtlib.reset_information()
    m = mutators_smtlib.IntroduceFreshVariable()
    assert isinstance(str(m), str)

    decl = [
        Node('declare-const', 'x', 'Bool'),
        Node('declare-const', 'y', ('_', 'BitVec', 8)),
        Node('declare-const', 'z', ('_', 'BitVec', 8)),
    ]
    smtlib.collect_information(decl)
    term = Node('not', 'x')
    exprs = [*decl, Node('assert', term)]
    assert not m.filter(Node('false'))
    assert not m.filter(Node('x'))
    assert m.filter(term)
    assert check_global_mutations(m, term, exprs, [[
        Node('declare-const', f'x{term.id}__fresh', 'Bool'), *decl,
        Node('assert', f'x{term.id}__fresh')
    ]])

    # actually, it is much more important to check that this mutator
    # does *not* apply for various cases, as it quickly leads to cycles.
    assert not m.filter(Node('false'))
    assert not m.filter(Node('1.0'))
    assert not m.filter(Node('_', 'bv123', '8'))
    assert not m.filter(Node('y'))
    assert not m.filter(Node(('_', 'zero_extend', '8'), 'y'))
    assert not m.filter(Node(('_', 'zero_extend', '0'), 'y'))
    assert m.filter(Node('bvor', 'y', 'z'))
    assert m.filter(Node('bvor', 'y', 'y'))


def test_smtlib_let_elimination():
    m = mutators_smtlib.LetElimination()
    assert isinstance(str(m), str)
    assert not m.filter(Node('and', 'x', 'y'))

    n = Node('let', (('a', 'x'), ('b', 'y')), ('and', 'a', 'b'))
    assert m.filter(n)
    assert check_mutations(m, n, [('and', 'a', 'b')])

    n = Node(
        'let',
        (('a', 'x'), ('b', 'y')),
    )
    assert m.filter(n)
    assert check_mutations(m, n, [])


def test_smtlib_let_substitution():
    m = mutators_smtlib.LetSubstitution()
    assert isinstance(str(m), str)
    assert not m.filter(Node('and', 'x', 'y'))
    n = Node('let', (('a', 'x'), ('b', 'y')), ('and', 'a', 'b'))
    assert m.filter(n)
    assert check_mutations(m, n, [('let', (('a', 'x'), ('b', 'y')),
                                   ('and', 'x', 'b')),
                                  ('let', (('a', 'x'), ('b', 'y')),
                                   ('and', 'a', 'y'))])
    n = Node('let', (('a', 'x'), ('b', 'y')), ('and', 'abc', 'def'))
    assert m.filter(n)
    assert check_mutations(m, n, [])

    n = Node('let', (('a', 'x'), ('b', 'y')))
    assert m.filter(n)
    assert check_mutations(m, n, [])


def test_smtlib_remove_annotation():
    m = mutators_smtlib.RemoveAnnotation()
    assert isinstance(str(m), str)

    assert not m.filter(Node('x'))
    assert not m.filter(Node('and', 'x', 'y'))
    assert m.filter(Node('!', 'x', ':foo'))
    assert m.filter(Node('!', ('and', 'x', 'y'), ':foo', 'bar'))
    assert check_mutations(m, Node('!', 'x', ':foo'), [('x')])
    assert check_mutations(m, Node('!', ('and', 'x', 'y'), ':foo', 'bar'),
                           [('and', 'x', 'y')])


def test_smtlib_simplify_logic():
    m = mutators_smtlib.SimplifyLogic()
    assert isinstance(str(m), str)

    assert not m.filter(Node('assert', 'x'))
    assert m.filter(Node('set-logic', 'QF_NRA'))
    assert check_mutations(m, Node('set-logic', 'QF_NRA'),
                           [('set-logic', 'QF_LRA')])
    assert check_mutations(m, Node('set-logic', 'QF_BVFPLRAS'), [
        ('set-logic', 'QF_FPLRAS'),
        ('set-logic', 'QF_BVLRAS'),
        ('set-logic', 'QF_BVFPLRA'),
        ('set-logic', 'QF_BVFPS'),
    ])


def test_smtlib_quoted_symbols():
    m = mutators_smtlib.SimplifyQuotedSymbols()
    assert isinstance(str(m), str)
    assert not m.filter(Node('x'))
    assert m.filter(Node('|x|'))
    assert not m.filter(Node('|"x|'))
    assert not m.filter(Node('and', 'x', 'y'))
    assert not m.filter(Node('and', '|x|', 'y'))
    assert check_mutations(m, Node('|x|'), ['x'])
    assert check_global_mutations(m, Node('|x|'), Node('assert', '|x|'),
                                  [Node('assert', 'x')])


def test_smtlib_simplify_symbol_names():
    m = mutators_smtlib.SimplifySymbolNames()
    assert isinstance(str(m), str)
    assert m.filter(Node('declare-const', 'x', 'Real'))
    assert not m.filter(Node('declare-const', 'false', 'Real'))
    assert m.filter(Node('declare-const', 'x', 'Real'))
    exprs = [
        Node('declare-const', 'abcdef', 'Bool'),
        Node('declare-datatype', 'Color', (('red', ), ('|green|', ))),
        Node('assert', 'abcdef'),
        Node('assert',
             ('exists', (('xy', 'Bool'), ('z', 'Bool')), ('and', 'xy', 'z'))),
    ]
    smtlib.collect_information(exprs)
    assert check_global_mutations(m, exprs[0], exprs, [
        apply_simp(exprs, Simplification({Node('abcdef'): Node('abc')}, [])),
        apply_simp(exprs, Simplification({Node('abcdef'): Node('abcde')}, [])),
        apply_simp(exprs, Simplification({Node('abcdef'): Node('bcdef')}, []))
    ])
    assert check_global_mutations(m, exprs[1], exprs, [
        apply_simp(exprs, Simplification({Node('Color'): Node('Co')}, [])),
        apply_simp(exprs, Simplification({Node('Color'): Node('Colo')}, [])),
        apply_simp(exprs, Simplification({Node('Color'): Node('olor')}, [])),
        apply_simp(exprs, Simplification({Node('red'): Node('re')}, [])),
        apply_simp(exprs, Simplification({Node('red'): Node('ed')}, [])),
        apply_simp(exprs, Simplification({Node('|green|'): Node('|gr|')}, [])),
        apply_simp(exprs, Simplification({Node('|green|'): Node('|gree|')},
                                         [])),
        apply_simp(exprs, Simplification({Node('|green|'): Node('|reen|')},
                                         []))
    ])
    assert check_global_mutations(m, exprs[3][1], exprs, [
        apply_simp(exprs, Simplification({Node('xy'): Node('x')}, [])),
        apply_simp(exprs, Simplification({Node('xy'): Node('y')}, []))
    ])



#
# ddSMT: A delta debugger for SMT benchmarks in SMT-Lib v2 format.
#
# This file is part of ddSMT.
#
# Copyright (C) 2013-2021 by the authors listed in AUTHORS file.
#
# ddSMT is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ddSMT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ddSMT.  If not, see <https://www.gnu.org/licenses/>.

import re

from .smtlib import *
from .mutator_utils import Simplification


class CheckSatAssuming:
    """Replace ``check-sat-assuming`` with a regular ``check-sat``.

    This discards any assumptions provided to ``check-sat-assuming``.
    """
    def filter(self, node):
        return node.has_ident() and node.get_ident() == 'check-sat-assuming'

    def mutations(self, node):
        return [Simplification({node.id: Node(Node('check-sat'))}, [])]

    def __str__(self):
        return 'substitute check-sat-assuming by check-sat'


class EliminateVariable:
    """Eliminate variables that occur as children of equalities.

    Globally replace every leaf node from an equality with every
    other node from the same equality. For example, for
    ``(= x (* a b) (+ c d))``, ``x`` is replaced with either
    ``(* a b)`` or ``(+ c d)``.
    """
    def filter(self, node):
        return is_eq(node) and len(node) > 1 and any(
            map(lambda n: n.is_leaf(), node[1:]))

    def global_mutations(self, node, input_):
        ops = node[1:]
        targets = list(filter(lambda n: n.is_leaf(), ops))
        for t in targets:
            for c in ops:
                if c == t:
                    continue
                if t in nodes.dfs(c):
                    # Avoid cycles (for example with core.ReplaceByChild)
                    continue
                substs = {}
                for n in nodes.dfs(input_):
                    if n == t and not is_definition_node(n):
                        substs[n.id] = c
                if substs:
                    yield Simplification(substs, [])

    def __str__(self):
        return 'eliminate variable from equality'


class InlineDefinedFuns:
    """Explicitly inline a defined function."""
    def filter(self, node):
        return is_defined_fun(node)

    def mutations(self, node):
        if node.id in map(lambda n: n.id, nodes.dfs(get_defined_fun(node))):
            # we are about to inline the function into its own body
            return []
        if is_definition_node(node):
            # we are about to inline the function into its own name
            return []
        res = get_defined_fun(node)
        if res == node:
            return []
        return [Simplification({node.id: res}, [])]

    def __str__(self):
        return 'inline defined function'


class IntroduceFreshVariable:
    """Replace a term by a fresh variable of the same sort."""
    def filter(self, node):
        # introducing a fresh variable may lead to cycles in various cases,
        # thus we exclude a variety of nodes here:
        # - constants, leaf nodes
        # - BV terms with a single variable and smaller bit-width.
        if is_const(node) or node.is_leaf() or is_definition_node(node):
            return False
        sort = get_sort(node)
        if sort is None:
            return False
        if is_bv_sort(sort):
            found = False
            for n in nodes.dfs(node):
                if is_var(n):
                    if found:
                        return True
                    found = True
                    bw = get_bv_width(n)
                    if bw > int(sort[2].data):
                        return True
            return False
        return True

    def global_mutations(self, node, input_):
        varname = Node(f'x{node.id}__fresh')
        if is_var(varname):
            return []
        var = Node('declare-const', varname, get_sort(node))
        return [Simplification({node.id: varname}, [var])]

    def __str__(self):
        return 'introduce fresh variable'


class LetElimination:
    """Substitute a ``let`` expression with its body."""
    def filter(self, node):
        return is_operator_app(node, 'let')

    def mutations(self, node):
        if len(node) <= 2:
            return []
        return [Simplification({node.id: node[2]}, [])]

    def __str__(self):
        return 'eliminate let binder'


class LetSubstitution:
    """Substitute a variable bound by a ``let`` binder into the nested term."""
    def filter(self, node):
        return is_operator_app(node, 'let')

    def mutations(self, node):
        if len(node) <= 2:
            return []
        for var in node[1]:
            if any(n == var[0] for n in nodes.dfs(node[2])):
                subs = nodes.substitute(node[2], {var[0]: var[1]})
                yield Simplification({node.id: Node(node[0], node[1], subs)},
                                     [])

    def __str__(self):
        return 'substitute variable into let body'


class RemoveAnnotation:
    """Remove an annotation ``(! t ...)`` from a term."""
    def filter(self, node):
        return is_operator_app(node, '!')

    def mutations(self, node):
        yield Simplification({node.id: node[1]}, [])

    def __str__(self):
        return 'remove annotation'


class SimplifyLogic:
    """Replace the logic specified in ``(check-logic ...)`` with a simpler one.
    """
    def filter(self, node):
        return node.has_ident() and node.get_ident() == 'set-logic'

    def mutations(self, node):
        logic = node[1]
        cands = []
        repls = {
            'BV': '',
            'FP': '',
            'UF': '',
            'S': '',
            'T': '',
            'NRA': 'LRA',
            'LRA': '',
            'NIA': 'LIA',
            'LIA': '',
            'NIRA': 'LIRA',
            'LIRA': 'LRA'
        }
        for r in repls:
            assert logic.is_leaf()
            if r in logic.data:
                cands.append(logic.data.replace(r, repls[r]))
        yield from [
            Simplification({node.id: Node('set-logic', c)}, []) for c in cands
        ]

    def __str__(self):
        return 'simplify logic'


class SimplifyQuotedSymbols:
    """Turn a quoted symbol into a simple symbol."""
    def filter(self, node):
        return is_piped_symbol(node) and re.match(
            '\\|[a-zA-Z0-9~!@$%^&*_+=<>.?/-]+\\|', node.data) is not None

    def mutations(self, node):
        return [Simplification({node.id: get_piped_symbol(node)}, [])]

    def global_mutations(self, node, input_):
        return [Simplification({node: get_piped_symbol(node)}, [])]

    def __str__(self):
        return 'simplify quoted symbol'


class SimplifySymbolNames:
    """Simplify variable names.

    This only has cosmetic impact except for cases where variable names
    are compared/ordered lexicographically (it may enable, for example,
    the application of mutator :class:`ddsmt.mutators_core.ReplaceByVariable`).
    """
    def filter(self, node):
        # check for is_const(node[1]) to avoid x -> false -> fals -> false
        # if the variable is irrelevant, false may be accepted by the solver
        return node.has_ident() and len(node) >= 2 and node.get_ident() in [
            'declare-const', 'declare-datatype', 'declare-datatypes',
            'declare-fun', 'declare-sort', 'define-fun', 'exists', 'forall'
        ] and not is_const(node[1])

    def global_mutations(self, node, input_):
        if node.get_ident() in ['declare-datatype', 'declare-datatypes']:
            for c in self.__flatten(Node(node[1:])):
                yield from self.__mutate_symbol(c, input_)
            return
        if node.get_ident() in ['exists', 'forall']:
            for v in node[1]:
                yield from self.__mutate_symbol(v[0], input_)
            return
        yield from self.__mutate_symbol(node[1], input_)

    def __flatten(self, n):
        """Yield given node as flattened sequence."""
        if n.is_leaf():
            yield n
        else:
            for e in n.data:
                yield from self.__flatten(e)

    def __mutate_symbol(self, symbol, input_):
        """Return a list of mutations of input_ based on simpler versions of
        symbol."""
        if is_piped_symbol(symbol):
            for s in self.__simpler(get_piped_symbol(symbol)):
                if not is_var(Node('|' + s + '|')):
                    yield Simplification({symbol: Node('|' + s + '|')}, [])
        else:
            for s in self.__simpler(symbol):
                if not is_var(Node(s)):
                    yield Simplification({symbol: Node(s)}, [])

    def __simpler(self, symbol):
        """Return a list of simpler versions of the given symbol."""
        sym = symbol.data
        if len(sym) > 3:
            yield sym[:len(sym) // 2]
        if len(sym) > 1:
            yield sym[:-1]
            yield sym[1:]

    def __str__(self):
        return 'simplify symbol name'


def get_mutators():
    """Returns a mapping from mutator class names to the name of their config
    options."""
    return {
        'CheckSatAssuming': 'check-sat-assuming',
        'EliminateVariable': 'eliminate-variables',
        'InlineDefinedFuns': 'inline-functions',
        'IntroduceFreshVariable': 'introduce-fresh-variables',
        'LetElimination': 'let-elimination',
        'LetSubstitution': 'let-substitution',
        'RemoveAnnotation': 'remove-annotation',
        'SimplifyLogic': 'simplify-logic',
        'SimplifyQuotedSymbols': 'simplify-quoted-symbols',
        'SimplifySymbolNames': 'simplify-symbol-names',
    }



#
# ddSMT: A delta debugger for SMT benchmarks in SMT-Lib v2 format.
#
# This file is part of ddSMT.
#
# Copyright (C) 2013-2021 by the authors listed in AUTHORS file.
#
# ddSMT is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ddSMT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ddSMT.  If not, see <https://www.gnu.org/licenses/>.

import logging
import re
import string

from . import nodes
from .nodes import Node

# Stores all declared or defined (first-order) constants with their sorts
__constants = {}
# Stores all defined functions with their return sorts
__defined_functions = {}
# Stores the ids of all nodes that are symbols within their definitions
# i.e. the id of x within ``(declare-const x Int)``
__definition_node_ids = set()
# Stores the sorts for all declared or defined symbols
__sort_lookup = {}
# Stores indices that should not be replaced by constants
__indices = set()
# Caches calls to get_sort
__get_sort_cache = {}
# Stores constants for datatype sorts
__datatypes_constants = {}
# Stores the sorts of datatype constructors
__datatypes_constructors = {}


def collect_information(exprs):  # noqa: C901
    """Initialize global lookups for first-order constants, defined functions
    and sorts of all these symbols."""
    global __constants
    global __defined_functions
    global __definition_node_ids
    global __sort_lookup
    global __indices
    global __datatypes_constants
    global __datatypes_constructors
    reset_information()

    for cmd in exprs:
        if not cmd.has_ident():
            continue
        name = cmd.get_ident()
        if name == 'declare-const':
            if not len(cmd) == 3:
                logging.trace(
                    f'Ignored command: "{cmd}" should have three children.')
                continue
            if not cmd[1].is_leaf():
                logging.trace(f'Ignored command: "{cmd[1]}" is not a leaf')
                continue
            __constants[cmd[1].data] = cmd[2]
            __definition_node_ids.add(cmd[1].id)
            __sort_lookup[cmd[1].data] = cmd[2]
        if name == 'declare-fun':
            if not len(cmd) == 4:
                logging.trace(
                    f'Ignored command: "{cmd}" should have four children.')
                continue
            if not cmd[1].is_leaf():
                logging.trace(f'Ignored command: "{cmd[1]}" is not a leaf')
                continue
            if cmd[2].is_leaf():
                logging.trace(f'Ignored command: "{cmd[2]}" is a leaf')
                continue
            if cmd[2] == tuple():
                __constants[cmd[1].data] = cmd[3]
            __definition_node_ids.add(cmd[1].id)
            __sort_lookup[cmd[1].data] = cmd[3]
        if name == 'define-fun':
            if not len(cmd) == 5:
                logging.trace(
                    f'Ignored command: "{cmd}" should have five children.')
                continue
            if not cmd[1].is_leaf():
                logging.trace(f'Ignored command: "{cmd[1]}" is not a leaf')
                continue
            if cmd[2].is_leaf():
                logging.trace(f'Ignored command: "{cmd[2]}" is a leaf')
                continue
            if cmd[2] == tuple():
                __constants[cmd[1]] = cmd[3]
            __defined_functions[cmd[1]] = (len(
                cmd[2]), lambda args, cmd=cmd: nodes.substitute(
                    cmd[4], {cmd[2][i][0]: args[i]
                             for i in range(len(args))}))
            __definition_node_ids.add(cmd[1].id)
            __definition_node_ids.add(cmd[4].id)
            __sort_lookup[cmd[1].data] = cmd[3]
        if name == 'declare-datatype':
            if not len(cmd) == 3:
                logging.trace(
                    f'Ignored command: "{cmd}" should have three children.')
                continue
            if cmd[2].is_leaf():
                logging.trace(f'Ignored command: "{cmd[2]}" is as leaf')
                continue
            sort = cmd[1]
            for constr in cmd[2]:
                __datatypes_constructors[constr[0]] = sort
                if len(constr) == 1:
                    __datatypes_constants.setdefault(sort, [])
                    __datatypes_constants[sort].append(constr[0])

        if name == 'declare-datatypes':
            if not len(cmd) == 3:
                logging.trace(
                    f'Ignored command: "{cmd}" should have three children.')
                continue
            if cmd[1].is_leaf() or cmd[2].is_leaf():
                logging.trace(f'Ignored command: "{cmd}" children are leafs')
                continue
            # we implicitly assume nullary sorts here
            if any(map(lambda n: n.is_leaf(), cmd[1])):
                logging.trace(
                    f'Ignore declare-datatypes because sort declarations can not be leaf nodes: {cmd[1]}'
                )
                continue
            sorts = [s[0] for s in cmd[1]]
            for id in range(len(sorts)):
                if id >= len(cmd[2]):
                    logging.trace(
                        f'Ignore "{sorts[id]}" as it lacks a constructor')
                    continue
                for constr in cmd[2][id]:
                    __datatypes_constructors[constr[0]] = sorts[id]
                    if len(constr) == 1:
                        __datatypes_constants.setdefault(sorts[id], [])
                        __datatypes_constants[sorts[id]].append(constr[0])

    # Collect additional term level information.
    for node in nodes.dfs(exprs):
        # Mark indices of indexed terms.
        if not node.is_leaf() and len(node) > 2 and node[0] == '_':
            for num in node[2:]:
                if num.data.isdigit():
                    __indices.add(num.id)
        # Determine sort of symbols introduced by let.
        if is_operator_app(node, 'let'):
            for var in node[1]:
                if len(var) != 2:
                    continue
                sym, term = var
                if sym.is_leaf():
                    __sort_lookup[sym.data] = get_sort(term)
                    __definition_node_ids.add(sym.id)
        # Determine sort of symbols introduced by quantifiers
        if is_operator_app(node, 'exists') or is_operator_app(node, 'forall'):
            for var in node[1]:
                if len(var) != 2:
                    continue
                sym, term = var
                if sym.is_leaf():
                    __sort_lookup[sym.data] = term
                    __definition_node_ids.add(sym.id)


def reset_information():
    """Reset global information gathered via collect_information.

    This is mainly to be used in the unit tests.
    """
    global __constants
    global __defined_functions
    global __definition_node_ids
    global __sort_lookup
    global __indices
    global __get_sort_cache
    global __datatypes_constants
    global __datatypes_constructors
    __constants = {}
    __defined_functions = {}
    __definition_node_ids = set()
    __sort_lookup = {}
    __indices = set()
    __get_sort_cache = {}
    __datatypes_constants = {}
    __datatypes_constructors = {}


# General utilities


def get_variables_with_sort(var_sort):
    """Return all variables with the sort ``var_sort``.

    Requires that global information has been populated via
    ``collect_information``.
    """
    return [v for v in __sort_lookup if __sort_lookup[v] == var_sort]


def introduce_variables(exprs, vars):
    """Adds new variables to a set of input expressions.

    Expects ``vars`` to contain declaration commands (like ``declare-
    fun``). Inserts the variables into ``exprs`` before the first SMT-
    LIB command that is not ``set-info`` or ``set-logic``. In particular,
    new variables are added before the current variable declarations.
    """
    pos = 0
    while pos < len(exprs):
        e = exprs[pos]
        if not e.has_ident():
            break
        if e.get_ident() not in ['set-info', 'set-logic']:
            break
        pos += 1
    return exprs[:pos] + vars + exprs[pos:]


# General semantic testers and testers


def is_index(node):
    global __indices
    return node.id in __indices


def is_leaf(node):
    """Check whether the ``node`` is a leaf node."""
    return node.is_leaf()


def is_var(node):
    """Return true if ``node`` is a variable (first order constant) node.

    Requires that global information has been populated via
    ``collect_information``.
    """
    return node.is_leaf() and node in __constants


def is_piped_symbol(node):
    """Checks whether the ``node`` is a quoted symbol."""
    return node.is_leaf() and node[0] == '|' and node[-1] == '|'


def get_piped_symbol(node):
    """Returns the actual symbol name from a quoted symbol ``node``."""
    assert is_piped_symbol(node)
    return Node(node[1:-1])


def is_operator_app(node, name):
    return node.has_ident() and node.get_ident() == name


def is_indexed_operator(node, name, index_count=1):
    """Return true if ``node`` is an indexed operator ``name`` and the given
    number of indices matches ``index_count``."""
    if node.is_leaf() or len(node) < 2:
        return False
    if node.has_ident() and node.get_ident() != '_':
        return False
    if node[1] != name:
        return False
    return len(node) == index_count + 2


def is_indexed_operator_app(node, name, index_count=1):
    return len(node) > 0 and is_indexed_operator(node[0], name, index_count)


def has_nary_operator(node):
    """Check whether ``node`` is an application of an n-ary operator."""
    if node.is_leaf() or not node.has_ident():
        return False
    return node.get_ident() in [
        '=>',
        'and',
        'or',
        'xor',
        '=',
        'distinct',
        '+',
        '-',
        '*',
        'div',
        '/',
        '<=',
        '<',
        '>=',
        '>',
        'bvand',
        'bvor',
        'bvadd',
        'bvmul',
        'concat',
        'fp.lt',
        'fp.gt',
        'fp.leq',
        'fp.geq',
    ]


def is_const(node):
    """Return true if ``node`` is a constant value."""
    return is_bool_const(node) or is_arith_const(node) or is_int_const(
        node) or is_real_const(node) or is_string_const(node) or is_bv_const(
            node) or is_fp_const(node)


def is_eq(node):
    """Checks whether ``node`` is an equality."""
    return node.has_ident() and node.get_ident() == '='


def get_default_constants(sort):
    """Return a list of default constants for the given ``sort``."""
    if sort == 'Bool':
        return [Node('false'), Node('true')]
    if sort == 'Int':
        return [Node('0'), Node('1')]
    if sort == 'Real':
        return [Node('0.0'), Node('1.0')]
    if is_bv_sort(sort):
        return [Node('_', c, sort[2]) for c in ['bv0', 'bv1']]
    if is_fp_sort(sort):
        if sort.is_leaf():
            if sort.data == 'Float16':
                ew = 5
                sw = 10
            elif sort.data == 'Float32':
                ew = 8
                sw = 23
            elif sort.data == 'Float64':
                ew = 11
                sw = 52
            else:
                assert sort.data == 'Float128'
                ew = 15
                sw = 112
        else:
            if isinstance(sort, Node):
                ew = int(sort[-2].data)
                sw = int(sort[-1].data) - 1
            else:
                ew = sort[-2]
                sw = sort[-1] - 1
        sign = Node('_', 'bv0', 1)
        signm = Node('_', 'bv1', 1)
        zero_ew = Node('_', 'bv0', ew)
        zero_sw = Node('_', 'bv0', sw)
        one_sw = Node('_', 'bv1', sw)
        ones_ew = Node('_', f'bv{2**ew - 1}', ew)
        return [
            # +zero
            Node('fp', sign, zero_ew, zero_sw),
            # -zero
            Node('fp', signm, zero_ew, zero_sw),
            # +NaN (note: only one possible NaN value is generated)
            Node('fp', sign, ones_ew, one_sw),
            # -NaN (note: only one possible NaN value is generated)
            Node('fp', signm, ones_ew, one_sw),
            # +oo
            Node('fp', sign, ones_ew, zero_sw),
            # -oo
            Node('fp', signm, ones_ew, zero_sw)
        ]
    if is_set_sort(sort):
        return [
            Node('as', 'emptyset', sort),
            *[Node('singleton', c) for c in get_default_constants(sort[1])]
        ]
    if sort in __datatypes_constants:
        return __datatypes_constants[sort]
    return []


def _get_sort_aux(node):  # noqa: C901
    """Get the sort of the given node (uncached).

    Return ``None`` if it can not be inferred. Requires that global
    information has been populated via ``collect_information``.
    """
    if node.is_leaf() and node.data in __sort_lookup:
        return __sort_lookup[node]
    if is_bool_const(node):
        return Node('Bool')
    if is_bv_const(node):
        return Node('_', 'BitVec', str(get_bv_width(node)))
    if is_int_const(node) and not is_index(node):
        return Node('Int')
    if is_real_const(node) and not is_index(node):
        return Node('Real')
    bvwidth = get_bv_width(node)
    # operators the return bit-vectors handled via get_bv_width
    if bvwidth != -1:
        return Node('_', 'BitVec', str(bvwidth))
    # non-indexed operators
    if node.has_ident() and len(node) > 1:
        if is_operator_app(node, 'ite') and len(node) > 2:
            return get_sort(node[2])
        ident = node.get_ident()
        # operators that return Bool
        if ident in [
                # core theory
                'not',
                '=>',
                'and',
                'or',
                'xor',
                '=',
                'distinct',
                # bv theory
                'bvult',
                'bvule',
                'bvugt',
                'bvuge',
                'bvslt',
                'bvsle',
                'bvsgt',
                'bvsge',
                # fp theory
                'fp.leq',
                'fp.lt',
                'fp.geq',
                'fp.gt',
                'fp.eq',
                'fp.isNormal',
                'fp.isSubnormal',
                'fp.isZero',
                'fp.isInfinite',
                'fp.isNaN',
                'fp.isNegative',
                'fp.isPositive',
                # int / real theory
                '<=',
                '<',
                '>>',
                '>',
                'is_int',
                # sets theory
                'member',
                'subset',
                # string theory
                'str.<',
                'str.in_re',
                'str.<=',
                'str.prefixof',
                'str.suffixof',
                'str.contains',
                'str.is_digit',
        ]:
            return Node('Bool')
        # operators that return Int
        if ident in [
                'div',
                'mod',
                'abs',
                'to_int',
                # string theory
                'str.len',
                'str.indexof',
                'str.to_code',
                'str.to_int',
                # sets theory
                'card'
        ]:
            return Node('Int')
        # operators that return Real
        if ident in ['/', 'to_real', 'fp.to_real']:
            return Node('Real')
        if ident in ['+', '-', '*']:
            if any(map(lambda n: get_sort(n) == 'Real', node[1:])):
                return Node('Real')
            elif get_sort(node[1]) == 'Int':
                return Node('Int')
            else:
                return None
        # operators that return floating-points
        if ident in [
                'fp.abs',
                'fp.max',
                'fp.min',
                'fp.neg',
                'fp.rem',
        ]:
            return get_sort(node[1])
        if ident in [
                'fp.add',
                'fp.div',
                'fp.fma',
                'fp.mul',
                'fp.roundToIntegral',
                'fp.sqrt',
                'fp.sub',
        ]:
            return get_sort(node[2])
        if ident == 'fp':
            ew = get_bv_width(node[2])
            sw = 1 + get_bv_width(node[3])
            return Node('_', 'FloatingPoint', ew, sw)
        if ident == 'select':
            asort = get_sort(node[1])
            if is_array_sort(asort):
                return asort[1]
            return None
        if ident == 'store':
            return get_sort(node[1])
        if ident in __datatypes_constructors:
            return __datatypes_constructors[ident]

    # indexed operators
    if is_indexed_operator_app(node, 'divisible'):
        return Node('Bool')
    if is_indexed_operator_app(node, 'to_fp', 2) \
       or is_indexed_operator_app(node, 'to_fp_unsigned', 2):
        idx = get_indices(node[0], node[0][1], 2)
        return Node('_', 'FloatingPoint', idx[0], idx[1])

    return None


def get_sort(node):
    """Get the sort of the given node (cached).

    Return ``None`` if it can not be inferred. Requires that global
    information has been populated via ``collect_information``.
    """
    global __get_sort_cache

    if node.id in __get_sort_cache:
        return __get_sort_cache[node.id]
    if node in __get_sort_cache:
        return __get_sort_cache[node]
    sort = _get_sort_aux(node)
    __get_sort_cache[node.id] = sort
    __get_sort_cache[node] = sort
    return sort


def get_indices(node, name, index_count=1):
    """Return a list with the indices of the given indexed operator."""
    assert is_indexed_operator(node, name, index_count)
    return [int(n.data) for n in node[2:]]


# Boolean


def is_bool_const(node):
    """Return true if ``node`` is a Boolean constant."""
    return node.is_leaf() and node.data in ['false', 'true']


# Arithmetic


def is_arith_const(node):
    """Return true if ``node`` is an arithmetic constant."""
    if node.has_ident() and node.get_ident() == '/' and len(node) == 3:
        return is_int_const(node[1]) and is_int_const(node[2])
    return (node.is_leaf()
            and re.match('[0-9]+(\\.[0-9]*)?$', node.data) is not None)


def is_int_const(node):
    """Return true if ``node`` is an int constant."""
    return node.is_leaf() and re.match('^[0-9]+$', node.data) is not None


def is_real_const(node):
    """Return true if ``node`` is a real constant."""
    if node.has_ident() and node.get_ident() == '/' and len(node) == 3:
        return is_int_const(node[1]) and is_int_const(node[2])
    return (node.is_leaf()
            and re.match('^[0-9]+(\\.[0-9]*)?$', node.data) is not None)


# Arrays


def is_array_sort(node):
    """Return true if ``node`` is a array sort."""
    if node is None or node.is_leaf() or len(node) != 3:
        return False
    if not node.has_ident() or node.get_ident() != 'Array':
        return False
    return True


# BV


def is_bv_sort(node):
    """Return true if ``node`` is a bit-vector sort."""
    if node is None or node.is_leaf() or len(node) != 3:
        return False
    if not node.has_ident() or node.get_ident() != '_':
        return False
    return node[1] == 'BitVec'


def is_bv_const(node):
    """Return true if ``node`` is a bit-vector constant."""
    if node.is_leaf():
        s = node.data
        if s.startswith('#b') and all(c in ['0', '1'] for c in s[2:]):
            return True
        if s.startswith('#x') and all(c in string.hexdigits for c in s[2:]):
            return True
        return False
    if len(node) != 3:
        return False
    if not node.has_ident() or node.get_ident() != '_':
        return False
    if not node.data[1].is_leaf():
        return False
    return node.data[1].data.startswith('bv')


def is_bv_comp(node):
    """Checks whether ``node`` is a bit-vector comparison."""
    return node.has_ident() and node.get_ident() == 'bvcomp'


def is_bv_not(node):
    """Checks whether ``node`` is a bit-vector bit-wise negation."""
    return node.has_ident() and node.get_ident() == 'bvnot'


def is_bv_neg(node):
    """Checks whether ``node`` is a bit-vector negation."""
    return node.has_ident() and node.get_ident() == 'bvneg'


def get_bv_width(node):  # noqa: C901
    """Return the bit-width of a bit-vector node.

    Asserts that ``node`` is a bit-vector node. Requires that global
    information has been populated via ``collect_information``.
    """
    if is_bv_const(node):
        if node.is_leaf():
            data = node.data
            if data.startswith('#b'):
                return len(data[2:])
            assert data.startswith('#x')
            return len(data[2:]) * 4
        return int(node[2].data)
    if node in __sort_lookup:
        bvsort = __sort_lookup[node]
        if is_bv_sort(bvsort):
            return int(bvsort[2].data)
        return -1
    if is_indexed_operator_app(node, 'zero_extend') \
       or is_indexed_operator_app(node, 'sign_extend'):
        return get_indices(node[0], node[0][1])[0] + get_bv_width(node[1])
    if is_indexed_operator_app(node, 'extract', 2):
        idx = get_indices(node[0], 'extract', 2)
        return idx[0] - idx[1] + 1
    if is_indexed_operator_app(node, 'repeat'):
        return get_indices(node[0], 'repeat')[0] * get_bv_width(node[1])
    if is_indexed_operator_app(node, 'rotate_left') \
       or is_indexed_operator_app(node, 'rotate_right'):
        return get_bv_width(node[1])
    if is_indexed_operator_app(node, 'fp.to_ubv') \
       or is_indexed_operator_app(node, 'fp.to_sbv'):
        return get_indices(node[0], node[0][1])[0]
    if node.has_ident():
        ident = node.get_ident()
        if ident in [
                'bvadd',
                'bvand',
                'bvashr',
                'bvmul',
                'bvnand',
                'bvneg',
                'bvnor',
                'bvnot',
                'bvor',
                'bvsdiv',
                'bvshl',
                'bvshr',
                'bvsmod',
                'bvsrem',
                'bvsub',
                'bvudiv',
                'bvurem',
                'bvxnor',
                'bvxor',
        ]:
            return get_bv_width(node[1])
        if ident == 'concat':
            assert len(node) == 3
            return get_bv_width(node[1]) + get_bv_width(node[2])
        if ident == 'bvcomp':
            return 1
        if ident == 'ite':
            bw = get_bv_width(node[2])
            if bw > 0:
                return bw
    return -1


def get_bv_constant_value(node):
    """Assume that node is a bit-vector constant and return ``(value, bit-
    width)``."""
    assert is_bv_const(node)
    if node.is_leaf():
        if node.data.startswith('#b'):
            return (int(node[2:], 2), len(node[2:]))
        assert node.data.startswith('#x')
        return (int(node[2:], 16), len(node[2:]) * 4)
    return (int(node[1][2:]), int(node[2].data))


# FP


def is_fp_sort(node):
    """Return true if ``node`` is a floating-point sort."""
    if is_leaf(node) \
       and node.data.startswith('Float') \
       and node[5:] in ['16', '32', '64', '128']:
        return True
    if not node.has_ident() or node.get_ident() != '_' or len(node) != 4:
        return False
    return node[1] == 'FloatingPoint'


def is_fp_const(node):
    """Return true if ``node`` is a floating-point constant."""
    return is_operator_app(node, 'fp')


# Functions


def is_definition_node(node):
    """Check whether ``node`` is the node that specifies the name of a symbol
    within the declaration of definition of this symbol.

    This would be the case for the node ``x`` within ``(declare-const x
    Int)``, but not for any other occurrence of ``x``.
    """
    return node.id in __definition_node_ids


def is_defined_fun(node):
    """Check whether ``node`` is a defined function.

    Requires that global information has been populated via
    ``collect_information``.
    """
    if isinstance(node, str) or node.is_leaf():
        return node in __defined_functions
    return node.has_ident() and node.get_ident() in __defined_functions


def get_defined_fun(node):
    """Return the defined function ``node``, instantiated with the arguments of
    ``node`` if necessary.

    Assumes ``is_defined_fun(node)``. Requires that global information
    has been populated via ``collect_information``.
    """
    assert is_defined_fun(node)
    if node.is_leaf():
        _, func = __defined_functions[node.data]
        return func([])
    arity, func = __defined_functions[node.get_ident()]
    if arity == len(node[1:]):
        return func(node[1:])
    return node


# Sets


def is_set_sort(node):
    """Return true if ``node`` is a set sort."""
    if node.is_leaf() or len(node) != 2:
        return False
    if not node.has_ident() or node.get_ident() != 'Set':
        return False
    return True


# Strings


def is_seq_type(node):
    """Checks whether the ``node`` is a seq type."""
    return is_operator_app(node, 'Seq')


def is_string_const(node):
    """Checks whether the ``node`` is a string constant."""
    return node.is_leaf() and node[0] == '"' and node[-1] == '"'



#! /usr/bin/env python3
#
# ddSMT: A delta debugger for SMT benchmarks in SMT-Lib v2 format.
#
# This file is part of ddSMT.
#
# Copyright (C) 2013-2021 by the authors listed in AUTHORS file.
#
# ddSMT is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ddSMT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ddSMT.  If not, see <https://www.gnu.org/licenses/>.

import logging
import os
import sys
import time

from . import checker
from . import strategy_ddmin
from . import strategy_hierarchical
from . import mutators
from . import nodeio
from . import nodes
from . import options
from . import debug_utils
from . import tmpfiles


class DDSMTException(Exception):
    def __init__(self, msg):
        self.__msg = msg

    def __str__(self):
        return "[ddsmt] Error: {}".format(self.__msg)


def check_options():
    # check input file
    if not os.path.isfile(options.args().infile):
        raise DDSMTException('input file is not a regular file')

    if options.args().parser_test:
        # only parse and print
        exprs = list(nodeio.parse_smtlib(open(options.args().infile).read()))
        print(nodeio.write_smtlib(sys.stdout, exprs))
        sys.exit(0)

    # check executable
    if not options.args().cmd:
        raise DDSMTException('No executable was specified as command')
    if not os.path.isfile(options.args().cmd[0]):
        raise DDSMTException('Command "{}" is not a regular file'.format(
            options.args().cmd[0]))
    if not os.access(options.args().cmd[0], os.X_OK):
        raise DDSMTException('Command "{}" is not executable'.format(
            options.args().cmd[0]))


def setup_logging():
    logging.CHAT = 25
    logging.addLevelName(logging.CHAT, "CHAT")
    logging.chat = lambda msg, *args, **kwargs: logging.log(
        logging.CHAT, msg, *args, **kwargs)
    logging.TRACE = 5
    logging.addLevelName(logging.TRACE, "TRACE")
    logging.trace = lambda msg, *args, **kwargs: logging.log(
        logging.TRACE, msg, *args, **kwargs)
    logging.basicConfig(format='[ddSMT %(levelname)s] %(message)s')
    verbositymap = {
        -2: logging.ERROR,
        -1: logging.WARN,
        0: logging.CHAT,
        1: logging.INFO,
        2: logging.DEBUG,
        3: logging.TRACE,
    }
    verbosity = options.args().verbosity - options.args().quietness
    logging.getLogger().setLevel(
        level=verbositymap.get(verbosity, logging.DEBUG))


def ddsmt_main():
    start_time_process = time.process_time()
    # general setup
    setup_logging()
    check_options()
    tmpfiles.init()

    with debug_utils.Profiler(True):
        # show what we are going to do
        logging.info("input file:   '{}'".format(options.args().infile))
        logging.info("output file:  '{}'".format(options.args().outfile))
        logging.info("command:      '{}'".format(" ".join(
            map(str,
                options.args().cmd))))
        if options.args().cmd_cc:
            logging.info("command (cc): '{}'".format(options.args().cmd_cc))

        # parse the input
        start_time = time.time()
        with open(options.args().infile, 'r') as infile:
            exprs = list(nodeio.parse_smtlib(infile.read()))
            nexprs = nodes.count_exprs(exprs)

        logging.debug("parsed {} s-expressions in {:.2f} seconds".format(
            nexprs,
            time.time() - start_time))

        if options.args().parser_test:
            nodes.write_smtlib_to_file(options.args().outfile, exprs)
            return

        # disable unused theories
        mutators.auto_detect_theories(exprs)
        # copy binaries to temp folder
        tmpfiles.copy_binaries()
        # perform golden runs to see what the solver is doing
        checker.do_golden_runs()

        orig_exprs = exprs
        # do the reduction
        if options.args().strategy in ('ddmin', 'hybrid'):
            exprs, ntests = strategy_ddmin.reduce(exprs)
        if options.args().strategy in ('hierarchical', 'hybrid'):
            exprs, ntests = strategy_hierarchical.reduce(exprs)
        end_time = time.time()

        # show the results
        if exprs != orig_exprs:
            ifilesize = os.path.getsize(options.args().infile)
            ofilesize = os.path.getsize(options.args().outfile)
            exprcount = nodes.count_exprs(exprs)
            proctime = time.process_time() - start_time_process
            sizeperc = ofilesize / ifilesize * 100
            exprperc = exprcount / nexprs * 100

            logging.info('')
            logging.info(f'runtime:         {end_time - start_time:.2f} s')
            logging.debug(f'main process:   {proctime:.2f} s')
            logging.info(f'tests:           {ntests}')
            logging.info('input file:')
            logging.info(f'  file size:     {ifilesize} B')
            logging.info(f'  s-expressions: {nexprs}')
            logging.info('reduced file:')
            logging.info(f'  file size:     {ofilesize} B ({sizeperc:3.1f}%)')
            logging.info(f'  s-expressions: {exprcount} ({exprperc:3.1f}%)')
        else:
            logging.warning('unable to minimize input file')



import pytest

from .. import mutators
from .. import options


def test_help(capsys):
    options.__PARSED_ARGS = None
    with pytest.raises(SystemExit):
        options.parse_options(mutators, ['--help'])
    captured = capsys.readouterr()
    assert ('usage: ddsmt [<options>] <infile> <outfile> <cmd> [<cmd options>]'
            in captured.out)


def test_help_all(capsys):
    options.__PARSED_ARGS = None
    with pytest.raises(SystemExit):
        options.parse_options(mutators, ['--help-all'])
    captured = capsys.readouterr()
    assert '--[no-]core' in captured.out


def test_dump_config(capsys):
    options.__PARSED_ARGS = None
    with pytest.raises(SystemExit):
        options.parse_options(mutators, ['--dump-config'])
    captured = capsys.readouterr()
    assert "'dump_config': True," in captured.out


def test_no_options(capsys):
    options.__PARSED_ARGS = None
    with pytest.raises(SystemExit):
        options.parse_options(mutators, [])
    captured = capsys.readouterr()
    assert ('error: the following arguments are required: infile, outfile, cmd'
            in captured.err)


def test_basic(capsys):
    options.__PARSED_ARGS = None
    options.parse_options(mutators, ['input.smt2', 'output.smt2', 'binary'])
    options.parse_options(mutators,
                          ['--no-core', 'input.smt2', 'output.smt2', 'binary'])
    options.parse_options(
        mutators,
        ['--no-core', '--constants', 'input.smt2', 'output.smt2', 'binary'])


def test_cmd_cc(capsys):
    options.__PARSED_ARGS = None
    options.parse_options(
        mutators,
        ['-c', 'other-binary', 'input.smt2', 'output.smt2', 'binary'])


def test_args(capsys):
    options.__PARSED_ARGS = None
    with pytest.raises(SystemExit):
        options.args()
    opts = options.args(['--no-core', 'input.smt2', 'output.smt2', 'binary'])
    assert options.args() == opts



import sys

from . import cli
from . import options
from . import debug_utils


def main():
    """Main entry point for ddsmt."""
    try:
        cli.ddsmt_main()
        if options.args().profile:
            debug_utils.render_profiles()
        return 0
    except MemoryError:
        print("[ddsmt] memory exhausted")
    except KeyboardInterrupt:
        print("[ddsmt] interrupted")
    except cli.DDSMTException as e:
        print(e)
    if options.args().profile:
        debug_utils.render_profiles()
    return 1


if __name__ == '__main__':
    sys.exit(main())
